import os
import numpy as np

from wntr.epanet.io import InpFile
from ptsnet.arrays import Table
from ptsnet.epanet.toolkit import ENepanet
from ptsnet.epanet.util import EN, FlowUnits, HydParam, to_si
from ptsnet.simulation.constants import G, TOL, FLOOR_FFACTOR, CEIL_FFACTOR, DEFAULT_FFACTOR
from ptsnet.simulation.constants import NODE_PROPERTIES, PIPE_PROPERTIES, PUMP_PROPERTIES, VALVE_PROPERTIES, OPEN_PROTECTION_PROPERTIES, CLOSED_PROTECTION_PROPERTIES
from ptsnet.simulation.validation import check_compatibility
from ptsnet.arrays.selectors import SelectorSet
from ptsnet.utils.data import imerge

from time import time

class Initializer:

    def __init__(self, inpfile, period = 0, skip_compatibility_check = False, warnings_on = True, _super = None):
        self.wn = get_water_network(inpfile)
        self.ng = self.wn.get_graph()
        self.ss = get_initial_conditions(inpfile, period = period, wn = self.wn)
        self.num_points = 0
        self.num_segments = 0
        self.where = None
        self._super = _super

        if not skip_compatibility_check:
            if warnings_on:
                t = time()
            try:
                check_compatibility(wn=self.wn, ss=self.ss)
            except Exception as e:
                if warnings_on:
                    print("Elapsed time (model check): ", time() - t, '[s]')
                raise e
            if warnings_on:
                print("Success - Compatible Model")
                print("Elapsed time (model check): ", time() - t, '[s]')

    def create_secondary_elements(self):
        # Open Protection Devices
        num_open_protections = len(self.ss['open_protection'])
        if num_open_protections > 0:
            open_protections = Table(OPEN_PROTECTION_PROPERTIES, num_open_protections)
            open_protection_labels = list(self.ss['open_protection'].keys())
            open_protections.assign_labels(open_protection_labels)
            for elem, props in self.ss['open_protection'].items():
                open_protections.node[elem] = props['node']
                open_protections.area[elem] = props['area']
            self.ss['open_protection'] = open_protections

        # Closed Protection Devices
        num_closed_protections = len(self.ss['closed_protection'])
        if num_closed_protections > 0:
            closed_protections = Table(CLOSED_PROTECTION_PROPERTIES, num_closed_protections)
            closed_protection_labels = list(self.ss['closed_protection'].keys())
            closed_protections.assign_labels(closed_protection_labels)
            for elem, props in self.ss['closed_protection'].items():
                closed_protections.node[elem] = props['node']
                closed_protections.area[elem] = props['area']
                closed_protections.height[elem] = props['height']
                closed_protections.water_level[elem] = props['water_level']
            self.ss['closed_protection'] = closed_protections

    def create_selectors(self):
        '''
        .points
            ['to_pipes']

            ['are_uboundaries']

            ['are_dboundaries']

            ['are_boundaries']

            ['start_valve']

            ['end_valve']

            ['single_valve']

            ['are_valve']

            ['start_pump']

            ['end_pump']

            ['single_pump']

            ['are_pump']

        .pipes
            ['to_nodes']

        .nodes
            ['not_in_pipes']

            ['in_pipes']

            ['to_points']

            ['to_points_are_uboundaries']

        .valves
            ['are_inline']

        '''
        self.where = SelectorSet(['points', 'pipes', 'nodes', 'valves', 'pumps', 'surge_protections'])

        self.where.pipes['to_nodes'] = imerge(self.ss['pipe'].start_node, self.ss['pipe'].end_node)
        pipes_idx = np.cumsum(self.ss['pipe'].segments+1).astype(int)
        self.where.points['to_pipes'] = np.zeros(pipes_idx[-1], dtype=int)
        for i in range(1, len(pipes_idx)):
            start = pipes_idx[i-1]
            end = pipes_idx[i]
            self.where.points['to_pipes'][start:end] = i
        self.where.points['are_uboundaries'] = np.cumsum(self.ss['pipe'].segments.astype(np.int)+1) - 1
        self.where.points['are_dboundaries'] = self.where.points['are_uboundaries'] - self.ss['pipe'].segments.astype(np.int)
        self.where.points['are_boundaries'] = imerge(self.where.points['are_dboundaries'], self.where.points['are_uboundaries'])
        self.where.points['are_boundaries',] = (np.arange(len(self.where.points['are_boundaries'])) % 2 != 0)

    def create_secondary_selectors(self):
        node_points_order = np.argsort(self.where.pipes['to_nodes'])
        self.where.nodes['not_in_pipes'] = np.isin(np.arange(self.wn.num_nodes),
            np.unique(np.concatenate((
                self.ss['valve'].start_node, self.ss['valve'].end_node,
                self.ss['pump'].start_node, self.ss['pump'].end_node)))
        )

        self.where.nodes['in_pipes'] = np.isin(self.where.pipes['to_nodes'], np.where(self.where.nodes['not_in_pipes'])[0])

        self.where.nodes['to_points'] = self.where.points['are_boundaries'][node_points_order]
        self.where.nodes['to_points_are_uboundaries'] = np.isin(self.where.nodes['to_points'], self.where.points['are_uboundaries'])
        self.where.nodes['to_points_are_dboundaries'] = np.isin(self.where.nodes['to_points'], self.where.points['are_dboundaries'])
        self.where.nodes['to_points',] = self.ss['node'].degree - self.where.nodes['not_in_pipes'] # Real degree

        # # Valve selectors
        self._create_nonpipe_selectors('valve')

        # # Pump selectors
        self._create_nonpipe_selectors('pump')

        # # Protection devices
        self._create_nonpipe_selectors('open_protection')
        self._create_nonpipe_selectors('closed_protection')

    def set_wave_speeds(self, default_wave_speed = 1000, wave_speed_file_path = None, delimiter = ',', wave_speed_method = 'optimal'):

            modified_lines = 0
            if wave_speed_file_path != None:
                with open(wave_speed_file_path, 'r') as f:
                    for line in f:
                        line = line.strip()
                        if len(line) <= 1:
                            raise ValueError("The wave_speed file has to have to entries per line 'pipe,wave_speed'")
                        pipe, wave_speed = line.split(delimiter)
                        self.ss['pipe'].wave_speed[pipe] = float(wave_speed)
                        modified_lines += 1

            if default_wave_speed != None:
                self.ss['pipe'].wave_speed[:] = default_wave_speed
                modified_lines = self.wn.num_pipes

            if modified_lines != self.wn.num_pipes:
                self.ss['pipe'].wave_speed[:] = 0
                excep = "The file does not specify wave speed values for all the pipes,\n"
                excep += "it is necessary to define a default wave speed value"
                raise ValueError(excep)

            self.ss['pipe'].desired_wave_speed = self.ss['pipe'].wave_speed
            self._set_segments(wave_speed_method)
            return True

    def _set_segments(self, wave_speed_method = 'optimal'):
        # method \in {'critical', 'user', 'dt', 'optimal'}
        if wave_speed_method in ('critical', 'dt', 'optimal'):
            self.ss['pipe'].segments = self.ss['pipe'].length
            self.ss['pipe'].segments /= self.ss['pipe'].wave_speed
            # Maximum time_step in the system to capture waves in all pipes
            max_dt = min(self.ss['pipe'].segments) / 2 # at least 2 segments in critical pipe
            self._super.settings.time_step = min(self._super.settings.time_step, max_dt)
            # The number of segments is defined
            self.ss['pipe'].segments /= self._super.settings.time_step
        elif wave_speed_method == 'user':
            self.ss['pipe'].segments = self.ss['pipe'].length
            self.ss['pipe'].segments /= (self.ss['pipe'].wave_speed * self._super.settings.time_step)
        elif wave_speed_method == 'dt':
            phi = self.ss['pipe'].length
        else:
            raise ValueError("Method is not compatible. Try: ['critical', 'user']")

        int_segments = np.round(self.ss['pipe'].segments)
        int_segments[int_segments < 2] = 2

        if wave_speed_method in ('critical', 'user'):
            # The wave_speed values are adjusted to compensate the truncation error
            self.ss['pipe'].wave_speed = self.ss['pipe'].wave_speed * self.ss['pipe'].segments/int_segments
            self.ss['pipe'].segments = int_segments
        elif wave_speed_method == 'dt':
            # The wave_speed is not adjusted when wave_speed_method == 'dt'
            # since the error is absorved by the time step
            self.ss['pipe'].segments = int_segments
        elif wave_speed_method == 'optimal':
            self.ss['pipe'].segments = int_segments
            phi = self.ss['pipe'].length / (self.ss['pipe'].wave_speed * self.ss['pipe'].segments)
            theta = np.dot(phi,np.ones_like(phi)) / np.dot(phi, phi)
            self._super.settings.time_step = 1/theta
            self.ss['pipe'].wave_speed = self.ss['pipe'].wave_speed * (phi*theta)
        self.ss['pipe'].dx = self.ss['pipe'].length / self.ss['pipe'].segments
        self.num_segments = int(sum(self.ss['pipe'].segments))
        self.num_points = self.num_segments + self.wn.num_pipes

    def _create_nonpipe_selectors(self, object_type):
        '''

        Notes:
        - Non-pipe selectors

        '''
        if object_type in ('valve', 'pump'):
            x1 = np.isin(self.where.pipes['to_nodes'], self.ss[object_type].start_node[self.ss[object_type].is_inline])
            x2 = np.isin(self.where.pipes['to_nodes'], self.ss[object_type].end_node[self.ss[object_type].is_inline])
            if object_type == 'valve':
                x3 = np.isin(self.where.pipes['to_nodes'], self.ss[object_type].start_node[~self.ss[object_type].is_inline])
            elif object_type == 'pump':
                x3 = np.isin(self.where.pipes['to_nodes'], self.ss[object_type].end_node[~self.ss[object_type].is_inline])
            self.where.points['start_' + object_type] = np.sort(self.where.points['are_boundaries'][x1])
            ordered = np.argsort(self.ss[object_type].start_node)
            ordered_end = np.argsort(self.ss[object_type].end_node)
            self.where.points['start_' + object_type,] = ordered[self.ss[object_type].is_inline[ordered]]
            last_order = np.argsort(self.where.points['start_' + object_type,])
            self.where.points['start_' + object_type][:] = self.where.points['start_' + object_type][last_order]
            self.where.points['start_' + object_type,][:] = self.where.points['start_' + object_type,][last_order]
            self.where.points['end_' + object_type] = np.sort(self.where.points['are_boundaries'][x2])
            self.where.points['end_' + object_type,] = ordered_end[self.ss[object_type].is_inline[ordered_end]]
            self.where.__dict__[object_type + 's']['are_inline'] = self.ss[object_type].is_inline
            self.where.__dict__[object_type + 's']['are_inline',] = last_order
            last_order = np.argsort(self.where.points['end_' + object_type,])
            self.where.points['end_' + object_type][:] = self.where.points['end_' + object_type][last_order]
            self.where.points['end_' + object_type,][:] = self.where.points['end_' + object_type,][last_order]
            self.where.points['single_' + object_type] = np.sort(self.where.points['are_boundaries'][x3])
            self.where.points['single_' + object_type,] = ordered[~self.ss[object_type].is_inline[ordered]]
            last_order = np.argsort(self.where.points['single_' + object_type,])
            self.where.points['single_' + object_type][:] = self.where.points['single_' + object_type][last_order]
            self.where.points['single_' + object_type,][:] = self.where.points['single_' + object_type,][last_order]
            self.where.points['are_' + object_type] = np.concatenate((
            self.where.points['single_' + object_type],
            self.where.points['start_' + object_type],
            self.where.points['end_' + object_type]))
            self.where.points['are_' + object_type].sort()
        elif object_type in ('open_protection', 'closed_protection'):
            protection_type = object_type[:object_type.find('_')]
            if self.ss[f'{protection_type}_protection']:
                self.where.nodes[f'are_{protection_type}_protection'] = self.ss[f'{protection_type}_protection'].node
                node_end_idx = np.cumsum(self.where.nodes['to_points',])[self.ss[f'{protection_type}_protection'].node] - 1
                node_start_idx = node_end_idx - self.where.nodes['to_points',][self.ss[f'{protection_type}_protection'].node] + 1
                self.where.points[f'start_{protection_type}_protection'] = self.where.nodes['to_points'][node_start_idx]
                self.where.points[f'end_{protection_type}_protection'] = self.where.nodes['to_points'][node_end_idx]
                self.where.points[f'are_{protection_type}_protection'] = imerge(self.where.points[f'start_{protection_type}_protection'], self.where.points[f'end_{protection_type}_protection'])

def get_water_network(inpfile):
    ENFile = InpFile()
    return ENFile.read(inpfile)

def get_initial_conditions(inpfile, period = 0, wn = None):

    # EPANET initialization

    file_prefix, _ = os.path.splitext(inpfile)
    rptfile = file_prefix + '.rpt'
    outfile = file_prefix + '.bin'

    if wn is None:
        wn = get_water_network(inpfile)

    network_graph = wn.get_graph()
    EPANET = ENepanet()
    EPANET.ENopen(inpfile, rptfile, outfile)
    EPANET.ENopenH()
    EPANET.ENinitH(0)

    # Data structures for node and link initial conditions
    nodes = Table(NODE_PROPERTIES, wn.num_nodes)
    node_labels = []
    pipes = Table(PIPE_PROPERTIES, wn.num_pipes)
    pipe_labels = []
    valves = Table(VALVE_PROPERTIES, wn.num_valves)
    valve_labels = []
    pumps = Table(PUMP_PROPERTIES, wn.num_pumps)
    pump_labels = []

    ss = {
        'node' : nodes,
        'pipe' : pipes,
        'valve' : valves,
        'pump' : pumps,
        'open_protection' : {},
        'closed_protection' : {}
    }

    # Run EPANET simulation
    t = 0
    while EPANET.ENnextH() > 0 and t <= period: # EPS
        EPANET.ENrunH()
        t += 1
    if t == 0: # Not EPS
        EPANET.ENrunH()

    flow_units = FlowUnits(EPANET.ENgetflowunits())

    # Retrieve node conditions
    for i in range(1, wn.num_nodes+1):
        node_id = EPANET.ENgetnodeid(i)
        node_labels.append(node_id)
        ss['node'].leak_coefficient[i-1] = EPANET.ENgetnodevalue(i, EN.EMITTER)
        ss['node'].demand[i-1] = EPANET.ENgetnodevalue(i, EN.DEMAND)
        ss['node'].head[i-1] = EPANET.ENgetnodevalue(i, EN.HEAD)
        ss['node'].pressure[i-1] = EPANET.ENgetnodevalue(i, EN.PRESSURE)
        ss['node'].type[i-1] = EPANET.ENgetnodetype(i)
        z = EPANET.ENgetnodevalue(i, EN.ELEVATION)
        if ss['node'].type[i-1] == EN.RESERVOIR:
            z = 0
        elif ss['node'].type[i-1] == EN.TANK:
            z = ss['node'].head[i-1] - ss['node'].pressure[i-1]
        ss['node'].elevation[i-1] = z
        ss['node'].degree[i-1] = network_graph.degree(node_id)

    # Unit conversion
    to_si(flow_units, ss['node'].leak_coefficient, HydParam.EmitterCoeff)
    to_si(flow_units, ss['node'].demand, HydParam.Flow)
    to_si(flow_units, ss['node'].head, HydParam.HydraulicHead)
    to_si(flow_units, ss['node'].pressure, HydParam.Pressure)
    to_si(flow_units, ss['node'].elevation, HydParam.Elevation)

    non_pipe_nodes = []
    p, pp, v = 0, 0, 0 # pipes, pumps, valves
    for i in range(1, wn.num_links+1):

        link = wn.get_link(EPANET.ENgetlinkid(i))
        ltype = link.link_type.lower()

        if link.link_type == 'Pipe':
            k = p; p += 1
        elif link.link_type == 'Pump':
            k = pp; pp += 1
        elif link.link_type == 'Valve':
            k = v; v += 1

        ss[ltype].start_node[k], ss[ltype].end_node[k] = EPANET.ENgetlinknodes(i)
        ss[ltype].flowrate[k] = EPANET.ENgetlinkvalue(i, EN.FLOW)
        ss[ltype].velocity[k] = EPANET.ENgetlinkvalue(i, EN.VELOCITY)

        # Indexes are adjusted to fit the new Table / Indexing in EPANET's C code starts in 1
        ss[ltype].start_node[k] -= 1
        ss[ltype].end_node[k] -= 1

        flow = to_si(flow_units, [ss[ltype].flowrate[k]], HydParam.Flow)[0]
        if abs(flow) < TOL:
            ss[ltype].direction[k] = 0
            ss[ltype].flowrate[k] = 0
            if link.link_type == 'Pipe':
                ss[ltype].ffactor[k] = DEFAULT_FFACTOR
        elif flow > TOL:
            ss[ltype].direction[k] = 1
        else:
            ss[ltype].direction[k] = -1
            ss[ltype].flowrate[k] *= -1
            ss[ltype].start_node[k], ss[ltype].end_node[k] = ss[ltype].end_node[k], ss[ltype].start_node[k]

        if ss['node'].degree[ss[ltype].start_node[k]] >= 2 and \
            ss['node'].degree[ss[ltype].end_node[k]] >= 2:
            ss[ltype].is_inline[k] = True

        if link.link_type in ('Pipe', 'Valve'):
            ss[ltype].diameter[k] = link.diameter
            ss[ltype].area[k] = np.pi * link.diameter ** 2 / 4
            ss[ltype].type[k] = EPANET.ENgetlinktype(i)
            ss[ltype].head_loss[k] = EPANET.ENgetlinkvalue(i, EN.HEADLOSS)

        if link.link_type == 'Pipe':
            pipe_labels.append(link.name)
            ss[ltype].length[k] = link.length
        elif link.link_type == 'Pump':
            pump_labels.append(link.name)
            ss[ltype].initial_status[k] = link.initial_status
            ss[ltype].setting[k] = ss[ltype].initial_status[k]
            ss[ltype].head_loss[k] = EPANET.ENgetlinkvalue(i, EN.HEADLOSS)
            # Pump curve parameters
            qp, hp = list(zip(*link.get_pump_curve().points)); qp = list(qp); hp = list(hp)
            qpp = to_si(flow_units, float(ss[ltype].flowrate[k]), HydParam.Flow)
            hpp = to_si(flow_units, float(ss[ltype].head_loss[k]), HydParam.HydraulicHead)
            qp.pop(); hp.pop(); qp.append(qpp); hp.append(abs(hpp))
            order = np.argsort(qp)
            qp = np.array(qp)[order]
            hp = np.array(hp)[order]
            ss[ltype].a2[k], ss[ltype].a1[k], ss[ltype].Hs[k] = np.polyfit(qp, hp, 2)
            # Source head
            ss[ltype].source_head[k] = ss['node'].head[ss[ltype].start_node[k]]
            non_pipe_nodes += [ss[ltype].start_node[k], ss[ltype].end_node[k]]
        elif link.link_type == 'Valve':
            valve_labels.append(link.name)
            ss[ltype].initial_status[k] = EPANET.ENgetlinkvalue(i, EN.INITSTATUS)
            ss[ltype].setting[k] = ss[ltype].initial_status[k]
            ss[ltype].flowrate[k] = to_si(flow_units, float(ss[ltype].flowrate[k]), HydParam.Flow)
            ha = ss['node'].head[ss[ltype].start_node[k]]
            hb = ss['node'].head[ss[ltype].end_node[k]] if ss['node'].degree[ss[ltype].end_node[k]] > 1 else 0
            hl = ha - hb
            if hl > 0:
                ss[ltype].K[k] = ss[ltype].flowrate[k]/(ss[ltype].area[k]*(2*G*hl)**0.5)
            non_pipe_nodes += [ss[ltype].start_node[k], ss[ltype].end_node[k]]

    EPANET.ENcloseH()
    EPANET.ENclose()

    # Unit conversion
    to_si(flow_units, ss['pipe'].head_loss, HydParam.HydraulicHead)
    to_si(flow_units, ss['pipe'].flowrate, HydParam.Flow)
    to_si(flow_units, ss['pump'].flowrate, HydParam.Flow)
    to_si(flow_units, ss['pipe'].velocity, HydParam.Velocity)
    to_si(flow_units, ss['pump'].velocity, HydParam.Velocity)
    to_si(flow_units, ss['pump'].head_loss, HydParam.HydraulicHead)
    to_si(flow_units, ss['valve'].head_loss, HydParam.HydraulicHead)
    to_si(flow_units, ss['valve'].velocity, HydParam.Velocity)

    idx = ss['pipe'].ffactor == 0
    ss['pipe'].ffactor[idx] = \
        (2*G*ss['pipe'].diameter[idx] * ss['pipe'].head_loss[idx]) \
            / (ss['pipe'].length[idx] * ss['pipe'].velocity[idx]**2)

    # ss['pipe'].ffactor[ss['pipe'].ffactor >= CEIL_FFACTOR] = CEIL_FFACTOR
    # ss['pipe'].ffactor[ss['pipe'].ffactor <= FLOOR_FFACTOR] = FLOOR_FFACTOR

    ss['valve'].curve_index.fill(-1)
    ss['pump'].curve_index.fill(-1)

    demanded = np.logical_and(ss['node'].type != EN.RESERVOIR, ss['node'].type != EN.TANK)
    KeKd = ss['node'].demand[demanded] / np.sqrt(ss['node'].pressure[demanded])
    ss['node'].demand_coefficient[demanded] = KeKd - ss['node'].leak_coefficient[demanded]

    nodes.assign_labels(node_labels)
    pipes.assign_labels(pipe_labels)
    valves.assign_labels(valve_labels)
    pumps.assign_labels(pump_labels)

    non_pipe_nodes = np.array(non_pipe_nodes)
    zero_flow_pipes = ss['pipe'].flowrate == 0
    zf = ss['pipe'].start_node[zero_flow_pipes]
    zf = np.concatenate((zf, ss['pipe'].end_node[zero_flow_pipes]))
    _fix_zero_flow_convention('valve', non_pipe_nodes, zf, wn, ss)
    _fix_zero_flow_convention('pump', non_pipe_nodes, zf, wn, ss)

    return ss

def _fix_zero_flow_convention(ltype, non_pipe_nodes, zf, wn, ss):
    # Define flow convention for zero flow pipes attached to
    zero_flow = np.where(np.isin(ss[ltype].start_node, zf))[0]
    for k in zero_flow:
        upipe = wn.get_links_for_node(ss['node'].ilabel(ss[ltype].start_node[k]))
        upipe.remove(ss[ltype].labels[k])
        upipe = ss['pipe'].lloc(upipe[0])
        dpipe = wn.get_links_for_node(ss['node'].ilabel(ss[ltype].end_node[k]))
        dpipe.remove(ss[ltype].labels[k])
        dpipe = ss['pipe'].lloc(dpipe[0])

        if ss['pipe'].end_node[upipe] != ss[ltype].start_node[k]:
            ss['pipe'].direction[upipe] = -1 if ss['pipe'].direction[upipe] != -1 else 1
            ss['pipe'].start_node[upipe], ss['pipe'].end_node[upipe] = ss['pipe'].end_node[upipe], ss['pipe'].start_node[upipe]
        else:
            if ss['pipe'].direction[upipe] == 0:
                ss['pipe'].direction[upipe] = 1

        if ss['pipe'].start_node[dpipe] != ss[ltype].end_node[k]:
            ss['pipe'].direction[dpipe] = -1 if ss['pipe'].direction[dpipe] != -1 else 1
            ss['pipe'].start_node[dpipe], ss['pipe'].end_node[dpipe] = ss['pipe'].end_node[dpipe], ss['pipe'].start_node[dpipe]
        else:
            if ss['pipe'].direction[dpipe] == 0:
                ss['pipe'].direction[dpipe] = 1


from ptsnet.simulation.constants import G
import numpy as np
from time import time
from scipy import optimize
from numba import jit

# ------------------ SIM STEPS ------------------
@jit(nopython = True)
def run_interior_step(Q0, H0, Q1, H1, B, R, Cp, Bp, Cm, Bm,
    has_plus, has_minus):
    """Solves flow and head for interior points

    All the numpy arrays are passed by reference,
    therefore, it is possible to override the
    values of the parameters of the function

    TODO: UPDATE ARGUMENTS
    """
    # Extreme points
    Cm[0] = H0[1] - B[0]*Q0[1]
    Cp[-1] = H0[-2] + B[-2]*Q0[-2]
    Bm[0] = B[0] + R[0]*abs(Q0[1])
    Bp[-1] = B[-2] + R[-2]*abs(Q0[-2])

    # The first and last nodes are skipped in the  loop considering
    # that they are boundary nodes (every interior node requires an
    # upstream and a downstream neighbor)
    for i in range(1, len(Q0)-1):
        Cm[i] = (H0[i+1] - B[i]*Q0[i+1]) * has_minus[i]
        Bm[i] = (B[i] + R[i]*abs(Q0[i+1])) * has_minus[i]
        Cp[i] = (H0[i-1] + B[i]*Q0[i-1]) * has_plus[i]
        Bp[i] = (B[i] + R[i]*abs(Q0[i-1])) * has_plus[i]
        H1[i] = (Cp[i]*Bm[i] + Cm[i]*Bp[i]) / (Bp[i] + Bm[i])
        Q1[i] = (Cp[i] - Cm[i]) / (Bp[i] + Bm[i])

def run_general_junction(H0, Q1, H1, E1, D1, Cp, Bp, Cm, Bm, Ke, Kd, Z, where):
    """Solves flow and head for boundary points attached to nodes

    All the numpy arrays are passed by reference,
    therefore, it is possible to override the
    values of the parameters of the function

    Arguments:
        TODO: UPDATE ARGUMENTS
    """
    Cm[where.points['jip_dboundaries']] /= Bm[where.points['jip_dboundaries']]
    Cp[where.points['jip_uboundaries']] /= Bp[where.points['jip_uboundaries']]
    Bm[where.points['jip_dboundaries']] = 1 / Bm[where.points['jip_dboundaries']]
    Bp[where.points['jip_uboundaries']] = 1 / Bp[where.points['jip_uboundaries']]

    sc = np.add.reduceat(Cm[where.points['just_in_pipes']] + Cp[where.points['just_in_pipes']]\
        , where.nodes['just_in_pipes',])
    sb = np.add.reduceat(Bm[where.points['just_in_pipes']] + Bp[where.points['just_in_pipes']]\
        , where.nodes['just_in_pipes',])

    X =  sc / sb
    K = ((Ke[where.nodes['all_just_in_pipes']] + Kd[where.nodes['all_just_in_pipes']])/sb)**2
    HZ = X - Z[where.nodes['all_just_in_pipes']]
    K[HZ < 0] = 0

    HH = ((2*X + K) - np.sqrt(K**2 + 4*K*HZ)) / 2
    H1[where.points['just_in_pipes']] = HH[where.points['just_in_pipes',]]
    H1[where.points['are_reservoirs']] = H0[where.points['are_reservoirs']]
    H1[where.points['are_tanks']] = H0[where.points['are_tanks']]
    Q1[where.points['jip_dboundaries']] = H1[where.points['jip_dboundaries']] \
        * Bm[where.points['jip_dboundaries']] - Cm[where.points['jip_dboundaries']]
    Q1[where.points['jip_uboundaries']] = Cp[where.points['jip_uboundaries']] \
        - H1[where.points['jip_uboundaries']] * Bp[where.points['jip_uboundaries']]

    # Get demand and leak flows
    HH -= Z[where.nodes['all_just_in_pipes']]
    HH[HH < 0] = 0 # No demand/leak flow with negative pressure
    num_jip = len(where.nodes['all_just_in_pipes'])
    E1[:num_jip] = Ke[where.nodes['all_just_in_pipes']] * np.sqrt(HH)
    D1[:num_jip] = Kd[where.nodes['all_just_in_pipes']] * np.sqrt(HH)

def run_valve_step(Q1, H1, Cp, Bp, Cm, Bm, setting, coeff, area, where):
    # --- End valves
    if len(where.points['single_valve',]) > 0:
        K0 = setting[where.points['single_valve',]] \
            * coeff[where.points['single_valve',]] \
                * area[where.points['single_valve',]] # C = K*sqrt(2g \Delta H)
        K = 2*G*(Bp[where.points['single_valve']] * K0)**2
        Cp_end = Cp[where.points['single_valve']]

        H1[where.points['single_valve']] =  \
            ((2*Cp_end + K) - np.sqrt((2*Cp_end + K)**2 - 4*Cp_end**2)) / 2

        Q1[where.points['single_valve']] = K0 * np.sqrt(2 * G * H1[where.points['single_valve']])

    # --- Inline valves
    if len(where.points['start_valve']) > 0:
        CM = Cm[where.points['end_valve']]
        BM = Bm[where.points['end_valve']]
        CP = Cp[where.points['start_valve']]
        BP = Bp[where.points['start_valve']]

        S = np.sign(CP - CM)
        CV = 2 * G * (setting[where.points['start_valve',]] \
            * coeff[where.points['start_valve',]] \
                * area[where.points['start_valve',]]) ** 2
        X = CV * (BP + BM)
        Q1[where.points['start_valve']] = (-S*X + S*np.sqrt(X**2 + S*4*CV*(CP - CM)))/2
        Q1[where.points['end_valve']] = Q1[where.points['start_valve']]
        H1[where.points['start_valve']] = CP - BP*Q1[where.points['start_valve']]
        H1[where.points['end_valve']] = CM + BM*Q1[where.points['start_valve']]

def run_pump_step(source_head, Q1, H1, Cp, Bp, Cm, Bm, a1, a2, Hs, setting, where):
    if len(where.points['single_pump']) > 0:
        CP = source_head[where.points['single_pump',]]
        BP = np.zeros(len(where.points['single_pump']))
        CM = Cm[where.points['single_pump']]
        BM = Bm[where.points['single_pump']]

        alpha = setting[where.points['single_pump',]]
        A = a2[where.points['single_pump',]]
        B = a1[where.points['single_pump',]]*alpha - BM - BP
        C = Hs[where.points['single_pump',]]*alpha**2 - CM + CP

        root = B ** 2 - 4 * A * C; root[root < 0] = 0
        Q = (-B - np.sqrt(root)) / (2*A); Q[Q < 0] = 0
        Q1[where.points['single_pump']] = Q

        hp = a2[where.points['single_pump',]] * Q**2 + a1[where.points['single_pump',]]*alpha * Q + \
            Hs[where.points['single_pump',]]*alpha**2

        H1[where.points['single_pump']] =  CP + hp

    if len(where.points['start_pump']) > 0:
        CP = Cp[where.points['start_pump']]
        BP = Bp[where.points['start_pump']]
        CM = Cm[where.points['end_pump']]
        BM = Bm[where.points['end_pump']]

        alpha = setting[where.points['start_pump',]]
        A = a2[where.points['start_pump',]]
        B = a1[where.points['start_pump',]]*alpha - BM - BP
        C = Hs[where.points['start_pump',]]*alpha**2 - CM + CP

        root = B ** 2 - 4 * A * C; root[root < 0] = 0
        Q = (-B - np.sqrt(root)) / (2*A); Q[Q < 0] = 0
        Q1[where.points['start_pump']] = Q
        Q1[where.points['end_pump']] = Q

        hp = a2[where.points['start_pump',]] * Q**2 + a1[where.points['start_pump',]]*alpha * Q + \
            Hs[where.points['start_pump',]]*alpha**2

        H1[where.points['start_pump']] =  CP - BP * Q
        H1[where.points['end_pump']] =  H1[where.points['start_pump']] + hp
        H1[where.points['end_pump']] =  CM + BM * Q1[where.points['end_pump']]

def tflow(QT1, QT0, CP, CM, BM, BP, HT0, tau, aT, VA0, C):
    CC = CP + CM # Cp/Bp + Cm/Bm
    BB = BM + BP # 1/Bp + 1/Bm
    Hb = 10.3 # Barometric pressure
    k = 0 # air-chamber orifice head loss coefficient
    m = 1.2 # gas exponent
    return ((CC - QT1)/BB + Hb - (HT0 + tau*(QT1+QT0)/(2*aT))-k*QT1*abs(QT1))* \
        (VA0 - aT*((QT1+QT0)*tau/(2*aT)))**m - C

def tflow_prime(QT1, QT0, CP, CM, BM, BP, HT0, tau, aT, VA0, C):
    CC = CP + CM # Cp/Bp + Cm/Bm
    BB = BM + BP # 1/Bp + 1/Bm
    Hb = 10.3 # Barometric pressure
    k = 0 # air-chamber orifice head loss coefficient
    m = 1.2 # gas exponent
    p1 = (-m*aT/(2 * aT/tau) * (VA0 - (QT0+QT1)*aT/(2 * aT/tau))**(m-1)* \
            ((CC-QT1)/BB + Hb - HT0 - (QT0+QT1)/(2 * aT/tau) - k*QT1*np.abs(QT1)))
    p2 = (-1/BB -1/(2 * aT/tau) - k*2.*QT1*np.sign(QT1)) * \
            (VA0 - (QT0+QT1)*aT/(2 * aT/tau))**m
    return p1+p2

def run_open_protections(H0, H1, Q1, QT, aT, Cp, Bp, Cm, Bm, tau, where):

    CP = Cp[where.points['start_open_protection']]
    BP = Bp[where.points['start_open_protection']]
    CM = Cm[where.points['end_open_protection']]
    BM = Bm[where.points['end_open_protection']]
    CC = CP + CM # Cp/Bp + Cm/Bm
    BB = BM + BP # 1/Bp + 1/Bm

    H1[where.points['start_open_protection']] = \
        (CC + QT + (2*aT*H0[where.points['start_open_protection']]/tau)) / (BB + (2*aT/tau))
    H1[where.points['end_open_protection']] = H1[where.points['start_open_protection']]
    Q1[where.points['start_open_protection']] = CP - H1[where.points['start_open_protection']]*BP
    Q1[where.points['end_open_protection']] = H1[where.points['end_open_protection']]*BM - CM
    QT = Q1[where.points['start_open_protection']] - Q1[where.points['end_open_protection']]

def run_closed_protections(H0, H1, Q1, QT0, QT1, HT0, HT1, VA, htank, aT, Cp, Bp, Cm, Bm, tau, C, where):

    CP = Cp[where.points['start_closed_protection']]
    BP = Bp[where.points['start_closed_protection']]
    CM = Cm[where.points['end_closed_protection']]
    BM = Bm[where.points['end_closed_protection']]
    CC = CP + CM # Cp/Bp + Cm/Bm
    BB = BM + BP # 1/Bp + 1/Bm

    QT1 = optimize.newton(
            tflow, QT0, fprime=tflow_prime,
        args =
            ( QT0, CP, CM, BM, BP, HT0, tau, aT, VA, C), tol=1e-10)

    HT1 = HT0 + (QT0+QT1)/(2 * aT/tau)
    H1[where.points['start_closed_protection']] = (CC - QT1)/BB
    H1[where.points['end_closed_protection']] = H1[where.points['start_closed_protection']]
    Q1[where.points['start_closed_protection']] = CP - H1[where.points['start_closed_protection']]*BP
    Q1[where.points['end_closed_protection']] = H1[where.points['end_closed_protection']]*BM - CM
    VA = (htank - HT1) * aT
    HT0 = HT1
    QT0 = QT1



import numpy as np

from scipy.interpolate import splev, splrep

def define_curve(X, Y):
    spl = splrep(X, Y)
    return lambda x : splev(x, spl)

def is_array(x):
    return (not type(x) in (str, np.str_)) and hasattr(x, "__iter__")

def imerge(A, B):
    d1 = len(A)
    d2 = len(B)

    if d1 != d2:
        raise ValueError("Both arrays should have the same dimensions")
    if len(A.shape) != len(B.shape) or len(A.shape) > 1:
        raise ValueError("Both arrays should be 1-dimensional")
    if A.dtype != B.dtype:
        raise ValueError("Both arrays should have the same dtype")

    x = np.zeros(2*d1, dtype = A.dtype)
    x[0::2] = A
    x[1::2] = B

    return x


import numpy as np

from ptsnet.epanet.util import EN
from ptsnet.simulation.constants import TOL

class ModelError(Exception):
    pass

def check_compatibility(wn, ss):

    min_degree = min(wn.get_graph().degree)
    if min_degree[1] == 0:
        raise ModelError("node '%s' is isolated" % min_degree[0])

    not_inline_pumps = np.arange(wn.num_pumps)[~ss['pump'].is_inline]
    inline_pumps = np.arange(wn.num_pumps)[ss['pump'].is_inline]
    inline_valves = np.arange(wn.num_valves)[ss['valve'].is_inline]

    # Pumps can not have dead ends
    possible_dead_end = ss['node'].degree[ss['pump'].end_node] < 2
    if possible_dead_end.any():
        raise ModelError("there are pumps with an incompatible end node" % ss['node'].labels[possible_dead_end])

    # Valves can not have isolated start nodes
    possible_isolated_vstart = ss['node'].degree[ss['valve'].start_node] < 2
    if possible_isolated_vstart.any():
        raise ModelError("there are valves with an incompatible start node" % ss['node'].labels[possible_isolated_vstart])

    # Pumps can not have isolated start nodes that are not reservoirs
    incompatible_start_pnodes = (ss['node'].degree[ss['pump'].start_node[not_inline_pumps]] == 1) * \
        (ss['node'].degree[ss['pump'].end_node[not_inline_pumps]] > 1)

    incompatible_start_pnodes = ~np.isin(
        ss['node'].type[ss['pump'].start_node[not_inline_pumps][incompatible_start_pnodes]], (EN.TANK, EN.RESERVOIR,))
    incompatible_start_pnodes = ss['node'].labels[not_inline_pumps][incompatible_start_pnodes]

    if len(incompatible_start_pnodes) > 0:
        raise ModelError("start nodes of pumps are incompatible: \n%s" % incompatible_start_pnodes)

    # Nodes of non-pipe elements can not be general junctions
    start_valve_error = ss['node'].degree[ss['valve'].start_node] > 2
    start_pump_error = ss['node'].degree[ss['pump'].start_node] > 2
    end_valve_error = ss['node'].degree[ss['valve'].end_node] > 2
    end_pump_error = ss['node'].degree[ss['pump'].end_node] > 2
    if start_valve_error.any():
        raise ModelError(
            "start nodes of valves are connected to more than one pipe: \n%s" % \
                ss['node'].labels[ss['valve'].start_node][start_valve_error])
    if start_pump_error.any():
        raise ModelError(
            "start nodes of pumps are connected to more than one pipe: \n%s" % \
                ss['node'].labels[ss['pump'].start_node][start_pump_error])
    if end_valve_error.any():
        raise ModelError(
            "end nodes of valves are connected to more than one pipe: \n%s" % \
                ss['node'].labels[ss['valve'].end_node][end_valve_error])
    if end_pump_error.any():
        raise ModelError(
            "end nodes of pumps are connected to more than one pipe: \n%s" % \
                ss['node'].labels[ss['pump'].end_node][end_pump_error])

    # Reservoirs can not be at the end of a pump
    if (ss['node'].type[ss['pump'].end_node] == EN.RESERVOIR).any():
        raise ModelError("there is a pump with a reservoir at its end node")
    # Reservoirs can not be at the end of a valve
    if (ss['node'].type[ss['valve'].end_node] == EN.RESERVOIR).any():
        raise ModelError("there is a valvve with a reservoir at its end node")

    all_non_pipe_nodes = np.concatenate((
        ss['pump'].start_node,
        ss['pump'].end_node,
        ss['valve'].start_node,
        ss['valve'].end_node))

    all_non_pipe_nodes_nburst = np.concatenate((
        ss['pump'].start_node,
        ss['pump'].end_node[inline_pumps],
        ss['valve'].start_node,
        ss['valve'].end_node[inline_valves]))

    # Non-pipe elements can only be connected to pipes
    if len(all_non_pipe_nodes) != len(np.unique(all_non_pipe_nodes)):
        raise ModelError("there are non-pipe elements connected to each other")

    # No leaks/demands are allowed for nodes of non-pipe elements (except end-valve)
    leaking = ss['node'].leak_coefficient[all_non_pipe_nodes_nburst] > 0
    demanded = ss['node'].demand_coefficient[all_non_pipe_nodes_nburst] > 0
    if leaking.any():
        raise ModelError("there is a non-pipe element connected to a leaking node\n%s" % \
            str(ss['node'].labels[all_non_pipe_nodes_nburst][leaking]))
    if demanded.any():
        raise ModelError("there are non-pipe elements connected to a node with demand\n%s" \
            % str(ss['node'].labels[all_non_pipe_nodes_nburst][demanded]))

    # Demands cannot be negative
    negative_demands = ss['node'].labels[np.where(ss['node'].demand < 0)[0]]
    negative_demands = negative_demands[np.where(~np.isin(negative_demands, wn.reservoir_name_list + wn.tank_name_list))[0]]
    if len(negative_demands) > 0:
        raise ModelError("nodes ['" + "', '".join(negative_demands) + "'] have negative demands")

    # Valves cannot have zero head_loss if initial flow is not zero
    non_zero_flow_valves = np.where((ss['valve'].flowrate != 0) & (ss['valve'].head_loss < TOL))[0]
    if len(non_zero_flow_valves) > 0:
        raise ModelError("valves ['" + "', '".join(ss['valve'].labels[non_zero_flow_valves]) + "'] cannot have zero (near-zero) loss coefficient")


from mpi4py import MPI

class CommManager:
    def __init__(self):
        self.MPI = MPI
        self.communicators = {}
        self.intra_communicators = {}
        self.add_communicator('main', MPI.COMM_WORLD)

    def __getitem__(self, index):
        if index in self.intra_communicators:
            return self.intra_communicators[index]
        elif index in self.communicators:
            return self.communicators[index]
        else:
            raise IndexError("(intra)communicator does not exist")

    def add_communicator(self, label, comm):
        if type(comm) == MPI.Intracomm:
            self.intra_communicators[label] = comm
        else:
            self.communicators[label] = comm


import os
import subprocess
import shutil

from pkg_resources import resource_filename

def run_shell(command):
    subprocess.run(command.split(' '))

def get_root_path():
    rpath = resource_filename(__name__, '')
    token = 'ptsnet'
    idx = rpath.rfind(token)
    return rpath[:idx+len(token)]

def get_examples_path():
    return os.path.join(get_root_path(), 'examples')

def get_example_path(example_name):
    ename = example_name
    if not example_name.lower().endswith('.inp'):
        ename = ename.upper()
        ename += '.inp'
    return os.path.join(get_examples_path(), ename)

def walk(folder_structure, root_path):
    paths = []

    if type(folder_structure) == dict:
        for ff in folder_structure:
            paths.extend(walk(folder_structure[ff], os.path.join(root_path, ff)))
    elif type(folder_structure) in (list, tuple):
        for ff in folder_structure:
            paths.extend(walk(ff, root_path))
    elif type(folder_structure) == int:
        new_root_path = os.path.join(root_path, str(folder_structure))
        paths.append(new_root_path)

    return paths

def get_workspaces():
    return os.listdir(os.path.join(get_root_path(), 'workspaces'))

def get_temp_folder():
    return os.path.join(get_root_path(), 'tmp')

def create_temp_folder():
    tmpdir = get_temp_folder()
    if os.path.exists(tmpdir): shutil.rmtree(tmpdir)
    os.makedirs(tmpdir)


import numpy as np

# ----- Global -----

WARNINGS = True
TIMEIT = True
PARALLEL = False
CEIL_FFACTOR = 10
FLOOR_FFACTOR = 1e-6
DEFAULT_FFACTOR = 0.035
G = 9.807 # SI gravity
TOL = 1E-6
COEFF_TOL = 1E-6
FILE_TEMPLATE = """
from ptsnet.simulation.sim import PTSNETSimulation
sim = PTSNETSimulation(workspace_id = {workspace_id}, inpfile = '{inpfile}', settings = {settings})\n
"""

# ----- Initial Conditions -----

NODE_PROPERTIES = {
    'demand' : float, #
    'head' : float, #
    'pressure' : float, #
    'elevation' : float, #
    'type' : int, #
    'degree' : int, #
    'processor' : int,
    # ----------------------------------------
    'leak_coefficient' : float, #
    'demand_coefficient' : float, #
}

PIPE_PROPERTIES = {
    'start_node' : int, #
    'end_node' : int, #
    'length' : float, #
    'diameter' : float, #
    'area' : float, #
    'wave_speed' : float, #
    'desired_wave_speed' : float, #
    'segments' : float, #
    'flowrate' : float, #
    'velocity' : float, #
    'head_loss' : float, #
    'direction' : int, #
    'ffactor' : float, #
    'dx' : float, #
    'type' : int, #
    'is_inline' : bool, #
}

PUMP_PROPERTIES = {
    'start_node' : int, #
    'end_node' : int, #
    'flowrate' : float, #
    'velocity' : float, #
    'head_loss' : float, #
    'direction' : int, #
    'initial_status' : int, #
    'is_inline' : bool, #
    'source_head' : float, #
    # ----------------------------------------
    'a1' : float, #
    'a2' : float, #
    'Hs' : float, #
    'curve_index' : int,
    'setting' : float,
}

VALVE_PROPERTIES = {
    'start_node' : int, #
    'end_node' : int, #
    'diameter' : float, #
    'area' : float, #
    'head_loss' : float, #
    'flowrate' : float, #
    'velocity' : float, #
    'direction' : int, #
    'initial_status' : int, #
    'type' : int, #
    'is_inline' : bool, #
    'adjustment' : float, #
    # ----------------------------------------
    'K' : float,
    'setting' : float,
    'curve_index' : int,
}

OPEN_PROTECTION_PROPERTIES = {
    'node' : int, #
    'area' : float, #
    # ----------------------------------------
    'QT' : float, # tank inflow
    'HT0' : float, # tank head t = 0
    'HT1' : float, # tank head t = 1
}

CLOSED_PROTECTION_PROPERTIES = {
    'node' : int, #
    'area' : float, #
    'height' : float,
    'water_level' : float,
    'C' : float, # ideal gas constant
    # ----------------------------------------
    'QT0' : float, # tank inflow t
    'QT1' : float, # tank inflow t + tau
    'HT0' : float, # tank head t
    'HT1' : float, # tank head t + tau
    'HA' : float, # air head
    'VA' : float, # air volume
}

SURGE_PROTECTION_TYPES = {
    'open' : 0,
    'closed' : 1
}

POINT_PROPERTIES = {
    'B' : float,
    'R' : float,
    'Bm' : float,
    'Bp' : float,
    'Cm' : float,
    'Cp' : float,
    'has_plus' : int,
    'has_minus' : int,
}

MEM_POOL_POINTS = {
    'flowrate' : float,
    'head' : float,
}

NODE_RESULTS = {
    'head' : float,
    'leak_flow' : float,
    'demand_flow' : float,
}

PIPE_START_RESULTS = {
    'flowrate' : float,
}

PIPE_END_RESULTS = {
    'flowrate' : float,
}

CLOSED_PROTECTION_RESULTS = {
    'water_level' : float
}

STEP_JOBS = (
    'run_step',
    'run_interior_step',
    'run_general_junction',
    'run_valve_step',
    'run_pump_step',
    'store_results',
)

INIT_JOBS = (
    'get_partition',
    '_create_selectors',
    '_define_dist_graph_comm',
    '_allocate_memory',
    '_load_initial_conditions'
)

COMM_JOBS = (
    'exchange_data',
    'barrier1',
    'barrier2'
)



from ptsnet.simulation.constants import G
import numpy as np
from time import time

# ------------------ SIM STEPS ------------------

def run_interior_step(Q0, H0, Q1, H1, B, R, Cp, Bp, Cm, Bm,
    has_plus, has_minus):
    """Solves flow and head for interior points

    All the numpy arrays are passed by reference,
    therefore, it is possible to override the
    values of the parameters of the function

    TODO: UPDATE ARGUMENTS
    """
    # Extreme points
    Cm[0] = H0[1] - B[0]*Q0[1]
    Cp[-1] = H0[-2] + B[-2]*Q0[-2]
    Bm[0] = B[0] + R[0]*abs(Q0[1])
    Bp[-1] = B[-2] + R[-2]*abs(Q0[-2])

    # The first and last nodes are skipped in the  loop considering
    # that they are boundary nodes (every interior node requires an
    # upstream and a downstream neighbor)
    for i in range(1, len(Q0)-1):
        Cm[i] = (H0[i+1] - B[i]*Q0[i+1]) * has_minus[i]
        Bm[i] = (B[i] + R[i]*abs(Q0[i+1])) * has_minus[i]
        Cp[i] = (H0[i-1] + B[i]*Q0[i-1]) * has_plus[i]
        Bp[i] = (B[i] + R[i]*abs(Q0[i-1])) * has_plus[i]
        H1[i] = (Cp[i]*Bm[i] + Cm[i]*Bp[i]) / (Bp[i] + Bm[i])
        Q1[i] = (Cp[i] - Cm[i]) / (Bp[i] + Bm[i])

def run_general_junction(H0, Q1, H1, E1, D1, Cp, Bp, Cm, Bm, Ke, Kd, Z, where):
    """Solves flow and head for boundary points attached to nodes

    All the numpy arrays are passed by reference,
    therefore, it is possible to override the
    values of the parameters of the function

    Arguments:
        TODO: UPDATE ARGUMENTS
    """
    for dboundary in where.points['jip_dboundaries']:
        Cm[dboundary] /= Bm[dboundary]
        Bm[dboundary] = 1 / Bm[dboundary]
    for uboundary in where.points['jip_uboundaries']:
        Cp[uboundary] /= Bp[uboundary]
        Bp[uboundary] = 1 / Bp[uboundary]

    X = np.zeros(len(where.nodes['just_in_pipes',]))
    sb = np.zeros(len(where.nodes['just_in_pipes',]))
    for i in range(len(where.nodes['just_in_pipes',])):
        j1 = where.nodes['just_in_pipes',][i]
        if i+1 == len(where.nodes['just_in_pipes',]):
            j2 = None
        else:
            j2 = where.nodes['just_in_pipes',][i+1]
        sc = sum(Cm[where.points['just_in_pipes']][j1:j2]) + \
                sum(Cp[where.points['just_in_pipes']][j1:j2])
        sb[i] = sum(Bm[where.points['just_in_pipes']][j1:j2]) + \
                sum(Bp[where.points['just_in_pipes']][j1:j2])
        X[i] =  sc / sb[i]

    HH = np.zeros(len(where.nodes['all_just_in_pipes']))
    for j, i in enumerate(where.nodes['all_just_in_pipes']):
        K = ((Ke[i] + Kd[i])/sb[j])**2
        HZ = X[j] - Z[i]
        if HZ < 0:
            K = 0
        HH[j] = ((2*X[j] + K) - np.sqrt(K**2 + 4*K*HZ)) / 2

    for j, i in enumerate(where.points['just_in_pipes',]):
        H1[where.points['just_in_pipes'][j]] = HH[i]
    for i in where.points['are_reservoirs']:
        H1[i] = H0[i]
    for i in where.points['are_tanks']:
        H1[i] = H0[i]
    for i in where.points['jip_dboundaries']:
        Q1[i] = H1[i] * Bm[i] - Cm[i]
    for i in where.points['jip_uboundaries']:
        Q1[i] = Cp[i] - H1[i] * Bp[i]

    # Get demand and leak flows
    for j, i in enumerate(where.nodes['all_just_in_pipes']):
        HH[j] -= Z[i]
        if HH[j] < 0:
            HH[j] = 0 # No demand/leak flow with negative pressure
        E1[j] = Ke[i] * np.sqrt(HH[j])
        D1[j] = Kd[i] * np.sqrt(HH[j])

def run_valve_step(Q1, H1, Cp, Bp, Cm, Bm, setting, coeff, area, where):
    # --- End valves
    ll = len(where.points['single_valve',])
    if ll > 0:
        K0 = np.zeros(ll)
        for i in where.points['single_valve',]:
            K0[i] = setting[i] * coeff[i] * area[i]
        for i, j in enumerate(where.points['single_valve']):
            K = 2*G*(Bp[j] * K0[i])**2
            H1[j] = ((2*Cp[j] + K) - np.sqrt((2*Cp[j] + K)**2 - 4*Cp[j]**2)) / 2
            Q1[j] = K0[i] * np.sqrt(2 * G * H1[j])

    # --- Inline valves
    ll = len(where.points['start_valve'])
    if ll > 0:
        for i, h in enumerate(where.points['start_valve',]):
            j = where.points['start_valve'][i]
            k = where.points['end_valve'][i]

            S = np.sign(Cp[j] - Cm[k])
            CV = 2 * G * (setting[h] * coeff[h] * area[h]) ** 2
            X = CV * (Bp[j] + Bm[k])
            Q1[j] = (-S*X + S*np.sqrt(X**2 + S*4*CV*(Cp[j] - Cm[k])))/2
            Q1[k] = Q1[j]
            H1[j] = Cp[j] - Bp[j]*Q1[j]
            H1[k] = Cm[k] + Bm[k]*Q1[j]

def run_pump_step(source_head, Q1, H1, Cp, Bp, Cm, Bm, a1, a2, Hs, setting, where):
    ll = len(where.points['single_pump'])
    if ll > 0:
        for i, j in enumerate(where.points['single_pump',]):
            k = where.points['single_pump'][i]
            BP = 0
            A = a2[j]
            B = a1[j]*setting[j] - Bm[k] - BP
            C = Hs[j]*setting[j]**2 - Cm[k] + source_head[j]

            root = B ** 2 - 4 * A * C
            if root < 0: root = 0
            Q = (-B - np.sqrt(root)) / (2*A)
            if Q < 0: Q = 0
            Q1[k] = Q

            hp = a2[j] * Q**2 + a1[j]*setting[j] * Q + Hs[j]*setting[j]**2

            H1[k] =  source_head[j] + hp

    ll = len(where.points['start_pump'])
    if ll > 0:
        for i, h in enumerate(where.points['start_pump',]):
            j = where.points['start_pump'][i]
            k = where.points['end_pump'][i]

            alpha = setting[h]
            A = a2[h]
            B = a1[h]*alpha - Bm[k] - Bp[j]
            C = Hs[h]*alpha**2 - Cm[k] + Cp[j]

            root = B ** 2 - 4 * A * C
            if root < 0: root = 0
            Q = (-B - np.sqrt(root)) / (2*A)
            if Q < 0: Q = 0
            Q1[j] = Q; Q1[k] = Q

            hp = a2[h] * Q**2 + a1[h]*alpha * Q + Hs[h]*alpha**2

            H1[j] =  Cp[j] - Bp[j] * Q
            H1[k] =  H1[j] + hp


import numpy as np

from collections import defaultdict as ddict
from ptsnet.utils.data import imerge
from pkg_resources import resource_filename

def even(num_points, num_processors):
    p = np.ones(num_points, dtype = int)
    n = num_points // num_processors
    r = num_points % num_processors
    for i in range(num_processors):
        start = i*n
        end = start + n
        if i < r:
            start += i; end += i
        elif r > 0:
            start += r; end += r
        p[start:end+1] = i
    return p

def _get_ghost_points(worker_points, worker_pipes):
    worker_points.sort()
    diff = np.where(np.diff(worker_pipes) > 0)[0]
    extradiff = []

    if len(diff) > 0:
        if worker_points[diff[0]] != worker_points[0]:
            extradiff.append(0)
        if worker_points[diff[-1]]+1 != worker_points[-1]:
            extradiff.append(-1)
    else:
        extradiff.append(0)
        extradiff.append(-1)

    # d is the list of indexes in worker_points that correspond to
    # ghost nodes that are assigned to the processor with id == rank
    ghosts = np.zeros(len(diff)*2 + len(extradiff), dtype=int)

    if len(extradiff) == 2:
        ghosts[1:-1] = imerge(diff, diff+1)
        ghosts[-1] = extradiff[-1]
    elif len(extradiff) == 1:
        if extradiff[0] == 0:
            ghosts[1:] = imerge(diff, diff+1)
        else:
            ghosts[0:-1] = imerge(diff, diff+1)
            ghosts[-1] = extradiff[-1]
    elif len(diff) > 0:
        ghosts[:] = imerge(diff, diff+1)
    elif len(diff) == 0:
        ghosts[:] = extradiff

    return ghosts

def get_partition(processors, rank, where, ss, wn, num_processors, inpfile):

    # List of points needs to be sorted
    worker_points = np.where(processors == rank)[0]
    worker_pipes = where.points['to_pipes'][worker_points]
    ghosts = _get_ghost_points(worker_points, worker_pipes)
    points_idx = np.ones_like(worker_points).astype(bool)
    points_idx[ghosts] = 0
    dependent = worker_points[ghosts]
    dependent_type = np.isin(dependent, where.points['are_uboundaries']) # 1: uboundary, 0: dboundary
    points = list(worker_points[points_idx])

    boundaries_to_nodes = {
        point : where.pipes['to_nodes'][i]
        for i, point in enumerate(where.points['are_boundaries'])
    }

    nodes = []; node_points_list = []; node_points_context = [0]
    tanks = []; tanks_points_list = []
    reservoirs = []; reservoirs_points_list = []
    inline_valves = []; start_valves = []; end_valves = []
    inline_pumps = []; start_pumps = [];  end_pumps = []
    single_valves = []; single_valve_points = []
    single_pumps = []; single_pump_points = []

    visited_nodes = []

    start_points = np.cumsum(where.nodes['to_points',])

    # Determine extra data for dependent points
    for i, b in enumerate(dependent):
        if b in boundaries_to_nodes: # Boundary point
            node = boundaries_to_nodes[b]
            degree = where.nodes['to_points',][node]

            nnode = wn.get_node(ss['node'].ilabel(node))
            l = wn.get_link(ss['pipe'].ilabel(where.points['to_pipes'][b]))
            snode = ss['pipe'].start_node[l.name]
            enode = ss['pipe'].end_node[l.name]

            if nnode.node_type in ('Tank', 'Reservoir'):
                if snode == node:
                    points.append(b+1)
                elif enode == node:
                    points.append(b-1)
                visited_nodes.append(node)
                points.append(b)
            if nnode.node_type == 'Tank':
                tanks.append(node)
                tanks_points_list.append(b)
                continue
            if nnode.node_type == 'Reservoir':
                reservoirs.append(node)
                reservoirs_points_list.append(b)
                continue

            if node in visited_nodes:
                continue

            if degree == 1:
                links = wn.get_links_for_node(nnode.name)
                if len(links) > 1:
                    l1 = wn.get_link(links[0])
                    l2 = wn.get_link(links[1])
                    nonpipe = l1 if l1.link_type.lower() != 'pipe' else l2
                    nonpipe_type = nonpipe.link_type.lower()
                    non_pipe_idx = ss[nonpipe_type].lloc(nonpipe.name)
                    nonpipe_start = ss[nonpipe_type].start_node[non_pipe_idx]
                    nonpipe_end = ss[nonpipe_type].end_node[non_pipe_idx]
                    is_inline = ss[nonpipe_type].is_inline[non_pipe_idx]

                    start_deg, end_deg = ss['node'].degree[[nonpipe_start, nonpipe_end]]
                    nonpipe_start_point = None
                    nonpipe_end_point = None

                    # Get upstream and downstream pipes
                    if start_deg == 2:
                        start_links = wn.get_links_for_node(ss['node'].ilabel(nonpipe_start))
                        start_links.remove(nonpipe.name)
                        start_pipe_idx = ss['pipe'].lloc(start_links[0])
                        if nonpipe_start == ss['pipe'].start_node[start_pipe_idx]:
                            npipe_spoint = 2*start_pipe_idx
                        elif nonpipe_start == ss['pipe'].end_node[start_pipe_idx]:
                            npipe_spoint = 2*start_pipe_idx + 1
                        nonpipe_start_point = where.points['are_boundaries'][npipe_spoint]
                    if end_deg == 2:
                        end_links = wn.get_links_for_node(ss['node'].ilabel(nonpipe_end))
                        end_links.remove(nonpipe.name)
                        end_pipe_idx = ss['pipe'].lloc(end_links[0])
                        if nonpipe_end == ss['pipe'].start_node[end_pipe_idx]:
                            npipe_epoint = 2*end_pipe_idx
                        elif nonpipe_end == ss['pipe'].end_node[end_pipe_idx]:
                            npipe_epoint = 2*end_pipe_idx + 1
                        nonpipe_end_point = where.points['are_boundaries'][npipe_epoint]
                    if is_inline:
                        nonpipe_points = [nonpipe_start_point, nonpipe_end_point]
                        processor_in_charge = min(processors[nonpipe_points])
                        processors[nonpipe_points] = processor_in_charge
                        points.append(b)
                        if processor_in_charge != rank:
                            visited_nodes.append(node)
                            continue
                        else:
                            points.append(nonpipe_start_point)
                            points.append(nonpipe_end_point)
                            if processors[nonpipe_start_point - 1] != rank:
                                points.append(nonpipe_start_point-1)
                            if processors[nonpipe_end_point + 1] != rank:
                                points.append(nonpipe_end_point+1)
                            if nonpipe_type == 'valve':
                                loc = ss['valve'].lloc(nonpipe.name)
                                if not loc in inline_valves:
                                    inline_valves.append(loc)
                                    start_valves.append(nonpipe_start_point)
                                    end_valves.append(nonpipe_end_point)
                            elif nonpipe_type == 'pump':
                                loc = ss['pump'].lloc(nonpipe.name)
                                if not loc in inline_pumps:
                                    inline_pumps.append(loc)
                                    start_pumps.append(nonpipe_start_point)
                                    end_pumps.append(nonpipe_end_point)
                            visited_nodes.append(node)
                            continue
                        continue
                    else:
                        points.append(b)
                        visited_nodes.append(node)
                        if nonpipe_type == 'valve':
                            if processors[b - 1] != rank:
                                points.append(b - 1)
                            single_valves.append(ss['valve'].lloc(nonpipe.name))
                            single_valve_points.append(b)
                            continue
                        if nonpipe_type == 'pump':
                            if processors[b + 1] != rank:
                                points.append(b + 1)
                            single_pumps.append(ss['pump'].lloc(nonpipe.name))
                            single_pump_points.append(b)
                            continue
                        continue

            start = 0 if node == 0 else start_points[node-1]
            end = start + where.nodes['to_points',][node]
            node_points = where.nodes['to_points'][start:end]
            processor_in_charge = min(processors[node_points])
            processors[node_points] = processor_in_charge

            if processor_in_charge != rank:
                if dependent_type[i] == 1: # uboundary
                    if processors[b-1] == rank:
                        points.append(b)
                elif dependent_type[i] == 0: # dboundary
                    if processors[b+1] == rank:
                        points.append(b)
                continue

            x = -1 if dependent_type[i] else 1

            if processors[b+x] != rank: # extra point needed
                points.append(b+x)

            to_points_are_uboundaries = where.nodes['to_points_are_uboundaries'][start:end]
            slots = np.arange(len(node_points)*2)
            idx = slots[slots % 2 == 0] + to_points_are_uboundaries
            slots[idx] = node_points
            pad = 1 - 2*to_points_are_uboundaries
            slots[idx + pad] = node_points + pad
            points += list(slots)
            nodes.append(node)
            node_points_list += list(node_points)
            node_points_context.append(len(node_points))
            visited_nodes.append(node)
        else:
            # Inner point
            node = None
            if b-1 in boundaries_to_nodes:
                node = boundaries_to_nodes[b-1]
                start = 0 if node == 0 else start_points[node-1]
                end = start + where.nodes['to_points',][node]
                node_points = where.nodes['to_points'][start:end]
                processor_in_charge = min(processors[node_points])
                processors[node_points] = processor_in_charge
            if b+1 in boundaries_to_nodes:
                node = boundaries_to_nodes[b+1]
                start = 0 if node == 0 else start_points[node-1]
                end = start + where.nodes['to_points',][node]
                node_points = where.nodes['to_points'][start:end]
                processor_in_charge = min(processors[node_points])
                processors[node_points] = processor_in_charge
            points.append(b-1); points.append(b); points.append(b+1)

    points = np.unique(points)
    points.sort()

    if len(points) == 0:
        partition = None
    else:
        partition = {
            'points' : {
                'global_idx' : points,
                'local_idx' : {p : i for i, p in enumerate(points)}
            },
            'nodes' : {
                'global_idx' : np.array(nodes).astype(int),
                'points' : np.array(node_points_list).astype(int),
                'context' : np.array(node_points_context).astype(int),
            },
            'tanks' : {
                'global_idx' : np.array(tanks).astype(int),
                'points' : np.array(tanks_points_list).astype(int),
            },
            'reservoirs' : {
                'global_idx' : np.array(reservoirs).astype(int),
                'points' : np.array(reservoirs_points_list).astype(int),
            },
            'inline_valves' : {
                'global_idx' : np.array(inline_valves).astype(int),
                'start_points' : np.array(start_valves).astype(int),
                'end_points' : np.array(end_valves).astype(int),
            },
            'inline_pumps' : {
                'global_idx' : np.array(inline_pumps).astype(int),
                'start_points' : np.array(start_pumps).astype(int),
                'end_points' : np.array(end_pumps).astype(int),
            },
            'single_valves' : {
                'global_idx' : np.array(single_valves).astype(int),
                'points' : np.array(single_valve_points).astype(int),
            },
            'single_pumps' : {
                'global_idx' : np.array(single_pumps).astype(int),
                'points' : np.array(single_pump_points).astype(int),
            }
        }

    return partition


import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import networkx.drawing.nx_pylab as nxp
import bisect, os, pickle
import numpy as np

from matplotlib.lines import Line2D
from ptsnet.utils.analytics import compute_wave_speed_error
from ptsnet.utils.io import get_temp_folder

def plot_wave_speed_error(sim, image_path):
    errors = compute_wave_speed_error(sim)
    intervals = np.linspace(0, 100, 5)
    percentile_labels = [("%.1f%%" + " - " + "%.1f%%") % (intervals[i], intervals[i+1],)  for i in range(len(intervals)-1)]
    error_intervals = {sim.ss['pipe'].labels[i] : bisect.bisect_left(intervals, errors[i]) for i in range(len(sim.ss['pipe'].labels))}
    colors = ['#4CD964', '#FFCC00', '#FF9500', '#FF3830']
    widths = [1, 1.5, 2, 2.5]
    custom_lines = [Line2D([0], [0], color = colors[i], lw = widths[i]) for i in range(len(widths))]
    start_nodes = sim.ss['node'].labels[sim.ss['pipe'].start_node]
    end_nodes = sim.ss['node'].labels[sim.ss['pipe'].end_node]
    G_pipes_only = list(zip(start_nodes, end_nodes))
    node_coords = {node_name : sim.wn.get_node(node_name).coordinates for node_name in sim.ss['node'].labels}
    G = sim.wn.get_graph()

    pipe_colors = [colors[error_intervals[pipe_name]-1] for pipe_name in sim.ss['pipe'].labels]
    pipe_widths = [widths[error_intervals[pipe_name]-1] for pipe_name in sim.ss['pipe'].labels]
    fig, ax = plt.subplots(figsize=(15,25))
    nxp.draw_networkx_nodes(G, node_coords, node_size = 0, label = None)
    nxp.draw_networkx_edges(G, node_coords, edgelist = G_pipes_only, edge_color = pipe_colors, width = pipe_widths, arrows = False)
    ax.legend(custom_lines, percentile_labels, title = 'Relative Error', fontsize = '15', title_fontsize = '17')
    plt.axis('off')
    fig.savefig(image_path)

def plot_times_per_step(duration=20):
    export_path = os.path.join(get_temp_folder(), "exported_times.pkl")
    if not os.path.exists(export_path):
        raise FileExistsError("There's no file with time results. You need to run ptsnet.utils.analytics.compute_simulation_times_per_step first")
    with open(export_path, 'rb') as f:
        data = pickle.load(f)

    time_steps = data['time_steps']
    processors = data['processors']
    n = len(time_steps)
    p = len(processors)
    num_steps = duration/np.tile(time_steps,(p,1)).T
    init_times = data['init_times']
    interior_times = data['interior_times']*num_steps
    boundary_times = data['boundary_times']*num_steps
    comm_times = data['comm_times']*num_steps
    totals = init_times + interior_times + boundary_times + comm_times

    matplotlib.rc('ytick', labelsize=22)
    loc = plticker.MultipleLocator(base=20e3)
    fig, ax = plt.subplots(figsize=(18, 8), dpi = 80)
    ax.yaxis.set_major_locator(loc)
    X = np.arange(p, dtype=int)
    max_y = 1.2*np.max(np.max(totals))
    plt.ylim(0, max_y)
    plt.yticks(np.linspace(0, max_y, 5))
    width = 0.8
    patterns = ['','','']
    bars1_1 = init_times + interior_times
    bars2_1 = bars1_1 + boundary_times

    for i in range(n-1,-1,-1):
        p1 = ax.bar(X-(n-i)*width/n, init_times[i], hatch=patterns[i], width = width/n, color = '#000', alpha = 1)
        p2 = ax.bar(X-(n-i)*width/n, interior_times[i], bottom = init_times[i], hatch=patterns[i], width = width/n, color = '#999', alpha = 1)
        p3 = ax.bar(X-(n-i)*width/n, boundary_times[i], bottom = bars1_1[i], hatch=patterns[i], width = width/n, color = '#ccc', alpha = 1)
        p4 = ax.bar(X-(n-i)*width/n, comm_times[i], bottom = bars2_1[i], hatch=patterns[i], width = width/n, color = '#eee', alpha = 1)

        for r1, r2, r3, r4 in zip(p1, p2, p3, p4):
            h1 = r1.get_height()
            h2 = r2.get_height()
            h3 = r3.get_height()
            h4 = r4.get_height()
            # plt.text(r1.get_x()+r1.get_width()/2., h1+h2+h3+h4, f'$\\tau_{i+1}$', ha = 'center', va='bottom', fontsize=22)
            plt.text(r1.get_x()+r1.get_width()/2., h1+h2+h3+h4, '{:,} s'.format(int(h1+h2+h3+h4)), ha = 'center', va='bottom', fontsize=16)
    plt.xticks(X+width/2-1, processors, fontsize = 22)
    box = ax.get_position()
    ax.set_position([box.x0, box.y0 + box.height * 0.1,
                    box.width, box.height * 0.9])
    plt.legend(
        (p1[0], p2[0], p3[0], p4[0],),
        (
            'Initialization',
            'Interior Points',
            'Boundary Conditions',
            'Communication'
        ),
        loc = 'upper center',
        bbox_to_anchor = (0.5, -0.15),
        fancybox = True,
        shadow = True,
        ncol = 4,
        fontsize = 18)

    plt.xlabel('Number of processors', fontsize = 20)
    ax.get_yaxis().set_major_formatter(
        matplotlib.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))
    plt.ylabel('Time [s]', fontsize = 20)
    plt.grid(True)
    fig.tight_layout()
    fig.subplots_adjust(top = 0.95, bottom = 0.2, left = 0.07)
    plt.savefig('times_bwsn.pdf')
    plt.show()


import setuptools
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ptsnet",
    version="0.0.19",
    author="Gerardo Riano",
    author_email="griano@utexas.edu",
    description="Parallel Transient Simulation in Water Networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gandresr/PTSNET",
    project_urls={
        "Bug Tracker": "https://github.com/gandresr/PTSNET/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.6"
)


from multiprocessing.sharedctypes import Value
from time import time
import numpy as np
import os, psutil, pickle

from ptsnet.utils.io import create_temp_folder, get_temp_folder
from ptsnet.simulation.constants import FILE_TEMPLATE
from ptsnet.simulation.sim import PTSNETSimulation, PTSNETSettings
from ptsnet.results.workspaces import num_workspaces

def compute_wave_speed_error(sim):
    ws = sim.ss['pipe'].wave_speed
    dws = sim.ss['pipe'].desired_wave_speed
    return np.abs(ws - dws)*100 / dws

def compute_num_processors(sim, plot=False, count=4, nprocessors=None, steps=2500):
    if type(sim) is not PTSNETSimulation: raise ValueError("'sim' must be a PTSNETSimulation")
    max_processors = psutil.cpu_count(logical=False) if nprocessors is None else nprocessors
    create_temp_folder()
    temp_file_path = os.path.join(get_temp_folder(), "temp_sim.py")
    compute_file_path = os.path.join(get_temp_folder(), "compute_processors_sim.py")

    # Modify settings for performance simulation only
    sim.settings.profiler_on = True
    sim.settings.duration = steps*sim.time_step
    sim.settings.save_results = True

    # Write Python script to run simulations
    with open(temp_file_path, "w") as f:
        f.write(FILE_TEMPLATE.format(workspace_id=None, inpfile=sim.inpfile, settings=sim.settings.to_dict(simplified=True)))
        f.write("sim.run()")
    cwd = os.getcwd()
    processors = np.linspace(1, max_processors, count, dtype=int)

    # Write bash executable
    bash_path = os.path.join(get_temp_folder(), "compute_num_processors.sh")
    with open(bash_path, "w") as f:
        i = 1
        f.write(f"echo 'Evaluating Performance (this might take a few minutes)'\n")
        for p in processors:
            f.write(f"mpiexec -n {p} python3 {temp_file_path} &> log.txt\n")
            f.write(f"echo '({int(100*i/len(processors))}%) Finished run {i}/{len(processors)}'\n")
            i += 1
        f.write(f"rm log.txt\n")
        f.write(f"python3 {compute_file_path}\n")

    # Write Python script to compute number of processors
    fcontent = "import matplotlib.pyplot as plt\n" if plot else ""
    fcontent += \
        "import numpy as np\n" + \
        "from pprint import pprint\n" + \
        "from ptsnet.simulation.sim import PTSNETSimulation\n" + \
        "from ptsnet.results.workspaces import num_workspaces\n" + \
        "from kneed import KneeLocator\n" + \
        f"workspaces = [num_workspaces()-1-i for i in range({count})]\n" + \
        "times = {}\n" + \
        "for w in workspaces:\n" + \
        "    with PTSNETSimulation(w) as sim:\n" + \
        "        exchange_data_time = np.mean(sim.profiler.summary['comm_jobs'].exchange_data)\n" + \
        "        barrier1_time = np.mean(sim.profiler.summary['comm_jobs'].barrier1)\n" + \
        "        barrier2_time = np.mean(sim.profiler.summary['comm_jobs'].barrier2)\n" + \
        "        run_step_time = np.mean(sim.profiler.summary['step_jobs'].run_step)\n" + \
        "        total_time = exchange_data_time + barrier1_time + barrier2_time + run_step_time\n" + \
        "        times[int(sim.settings.num_processors)] = float(total_time)\n" + \
        "x = list(times.keys()); x.sort()\n" + \
        "y = [times[i] for i in x]\n" + \
        "optimal = -1; kl = None\n" + \
        "if y[1] > y[0]:\n" + \
        "    optimal = 1\n" + \
        "else:\n" + \
        "    try:\n" + \
        "        kl = KneeLocator(x, y, curve='convex', direction='decreasing')\n" + \
        "        optimal = kl.knee\n" + \
        "    except:\n" + \
        "        pass\n" + \
        "    if optimal == -1: optimal = x[-1]\n" + \
        "print(f'\\nProcessor Times: \\n')\n" + \
        "pprint(times)\n" + \
        "print(f'\\n--> Recommended number of processors: {optimal}\\n')\n"

    if plot:
        fcontent += \
        "plt.plot(x, y, '-o')\n" + \
        "plt.axvline(x = optimal)\n" + \
        "plt.xlabel('Number of processors')\n" + \
        "plt.ylabel('Time [s]')\n" + \
        "plt.title('Average Time per Step')\n" + \
        "plt.savefig('knee.pdf')"

    with open(compute_file_path, "w") as f:
        f.write(fcontent)

    print("\nExecute the following command on your terminal:")
    print(f"bash {bash_path}\n")
    exit()

def compute_simulation_times_per_step(inpfile, time_steps, duration=20, plot=False, count=4, nprocessors=None, steps=2500):
    create_temp_folder()
    max_processors = psutil.cpu_count(logical=False) if nprocessors is None else nprocessors
    processors = np.linspace(1, max_processors, count, dtype=int)
    compute_file_path = os.path.join(get_temp_folder(), "compute_times_per_step.py")

    ii = 0
    for ts in time_steps:
        temp_file_path = os.path.join(get_temp_folder(), f"temp_sim_{ii}.py")
        # Write Python script to run simulations
        settings = PTSNETSettings(
            time_step = ts,
            duration = steps*ts,
            warnings_on = False,
            skip_compatibility_check = True,
            save_results = True,
            profiler_on = True,
            wave_speed_method = 'user'
        )
        with open(temp_file_path, "w") as f:
            f.write(FILE_TEMPLATE.format(workspace_id=None, inpfile=inpfile, settings=settings.to_dict(simplified=True)))
            f.write("sim.run()")
        cwd = os.getcwd()
        ii += 1

    sims = {(processors[ii], time_steps[jj]) : (ii, jj)  for ii in range(len(processors)) for jj in range(len(time_steps))}

    # Write bash executable
    bash_path = os.path.join(get_temp_folder(), "compute_times_per_step.sh")
    with open(bash_path, "w") as f:
        jj = 1
        f.write(f"echo 'Executing Simulations to Evaluate Performance'\n")
        f.write(f"echo '[This might take a few minutes]'\n")
        for ii, (p, ts) in enumerate(sims):
            temp_file_path = os.path.join(get_temp_folder(), f"temp_sim_{ii%len(time_steps)}.py")
            f.write(f"mpiexec -n {p} python3 {temp_file_path} &> log.txt\n")
            f.write(f"echo '({int(100*(1+ii)/len(sims))}%) Finished Simulation {1+ii}/{len(sims)} -> {p} processor(s) | time_step = {ts} s'\n")
            jj += 1
        f.write(f"rm log.txt\n")
        f.write(f"python3 {compute_file_path}\n")

    # Write Python script to compute simulation times
    export_path = os.path.join(get_temp_folder(), "exported_times.pkl")
    fcontent = \
    "import numpy as np\n" + \
    "import pickle\n" + \
    "from ptsnet.simulation.sim import PTSNETSimulation\n" + \
    "from ptsnet.results.workspaces import num_workspaces\n" + \
    f"sims = {sims}\n" + \
    f"processors = {str(list(processors))}\n" + \
    f"ntsteps = {len(time_steps)}; nproc = {count}\n" + \
    "init_times = np.zeros((ntsteps, nproc,))\n" + \
    "interior_times = np.zeros((ntsteps, nproc,))\n" + \
    "boundary_times = np.zeros((ntsteps, nproc,))\n" + \
    "comm_times = np.zeros((ntsteps, nproc,))\n" + \
    "totals = np.zeros((ntsteps, nproc,))\n" + \
    f"time_steps = {str(list(time_steps))}\n" + \
    "for ii, (p, ts) in enumerate(sims):\n" + \
    f"    jj = time_steps.index(ts) # index for time step\n" + \
    "    kk = processors.index(p)\n" + \
    "    workspace_id = num_workspaces() - len(sims) + ii\n" + \
    "    with PTSNETSimulation(workspace_id) as sim:\n" + \
    "        init_times[jj, kk] = \\\n" + \
    "            (sim.profiler.summary['get_partition'] + \\\n" + \
    "            sim.profiler.summary['_create_selectors'] + \\\n" + \
    "            sim.profiler.summary['_define_dist_graph_comm'] + \\\n" + \
    "            sim.profiler.summary['_allocate_memory'] + \\\n" + \
    "            sim.profiler.summary['_load_initial_conditions'])\n" + \
    "        interior_times[jj,kk] = \\\n" + \
    "            sim.profiler.summary['run_interior_step'] / sim.settings.time_steps\n" + \
    "        boundary_times[jj,kk] = \\\n" + \
    "            (sim.profiler.summary['run_general_junction'] + \\\n" + \
    "            sim.profiler.summary['run_valve_step'] + \\\n" + \
    "            sim.profiler.summary['run_pump_step']) / sim.settings.time_steps\n" + \
    "        comm_times[jj,kk] = \\\n" + \
    "            (sim.profiler.summary['exchange_data'] + \\\n" + \
    "            sim.profiler.summary['barrier1'] + \\\n" + \
    "            sim.profiler.summary['barrier2']) / sim.settings.time_steps\n" + \
    "        totals = init_times + interior_times + boundary_times + comm_times\n" + \
    "    times = {\n" + \
    "        'init_times' : init_times,\n" + \
    "        'interior_times' : interior_times,\n" + \
    "        'boundary_times' : boundary_times,\n" + \
    "        'comm_times' : comm_times,\n" + \
    "        'totals' : totals,\n" + \
    f"        'processors' : processors,\n" + \
    f"        'time_steps' : time_steps\n" + \
    "    }\n" + \
    f"    with open('{export_path}', 'wb') as f:\n" + \
    "        pickle.dump(times, f)\n"
    if plot:
        fcontent += \
        "# -----------------------------------------------------\n" + \
        "# | Execute after running the bash generated by\n" + \
        "# | compute_simulation_times_per_step\n" + \
        "# v\n" + \
        "from ptsnet.graphics.static import plot_times_per_step\n" + \
        f"plot_times_per_step(duration={duration})\n"

    with open(compute_file_path, "w") as f:
        f.write(fcontent)

    print("\nExecute the following command on your terminal:")
    print(f"bash {bash_path}\n")
    exit()


import numpy as np

from collections import defaultdict as ddict
from ptsnet.arrays import Table2D, Table, ObjArray
from ptsnet.parallel.partitioning import even, get_partition
from ptsnet.simulation.constants import MEM_POOL_POINTS, PIPE_START_RESULTS, PIPE_END_RESULTS, NODE_RESULTS, CLOSED_PROTECTION_RESULTS, POINT_PROPERTIES, G
from ptsnet.utils.data import is_array
from ptsnet.arrays.selectors import SelectorSet
from ptsnet.simulation.funcs import run_general_junction, run_interior_step, run_pump_step, run_valve_step, run_open_protections, run_closed_protections
from ptsnet.profiler.profiler import Profiler

class Worker:
    def __init__(self, **kwargs):
        self.send_queue = None
        self.recv_queue = None
        self.router = kwargs['router']
        self.wn = kwargs['wn']
        self.ss = kwargs['ss']
        self.global_where = kwargs['where']
        self.time_step = kwargs['time_step']
        self.time_steps = kwargs['time_steps']
        self.mem_pool_points = None
        self.point_properties = None
        self.num_nodes = 0 # number of nodes in worker
        self.num_start_pipes = 0 # number of start pipes in worker
        self.num_end_pipes = 0 # number of end pipes in worker
        self.num_jip_nodes = 0 # number of just-in-pipes junction nodes in worker
        self.num_open_protections = 0 # number of open surge protections
        self.num_closed_protections = 0 # number of closed surge protections
        self.where = SelectorSet(['points', 'pipes', 'nodes', 'valves', 'pumps', 'open_protections', 'closed_protections'])
        self.processors = even(kwargs['num_points'], self.router['main'].size)
        self.is_innactive = False
        innactive_processors = np.empty(self.router['main'].size, dtype=bool)
        self.results = {}
        self.profiler_on = kwargs['profiler_on']
        self.profiler = Profiler(self.router['main'].rank, is_on = self.profiler_on)

        ###
        self.profiler.start('get_partition')
        self.partition = get_partition(
            self.processors, self.router['main'].rank, self.global_where, self.ss,
            self.wn, self.router['main'].size, kwargs['inpfile'])
        self.profiler.stop('get_partition')
        ###

        ###
        self.profiler.start('check_processor_innactivity')
        if self.partition is None:
            self.is_innactive = True
        self.router['main'].Alltoall(np.ones(self.router['main'].size, dtype=bool)*self.is_innactive, innactive_processors)
        if np.any(innactive_processors):
            self.is_innactive = True
            self.profiler.stop('check_processor_innactivity')
            ###
            raise SystemError(" Partitioning is innecficient due to unused processor(s), try executing the parallel routine with less processors")
        self.profiler.stop('check_processor_innactivity')
        ###

        self.points = self.partition['points']['global_idx']
        self.num_points = len(self.points) # ponts assigned to the worker
        self.local_points = np.arange(self.num_points)

        ###
        self.profiler.start('_create_selectors')
        self._create_selectors()
        self.profiler.stop('_create_selectors')
        ###

        ###
        self.profiler.start('_define_worker_comm_queues')
        self._define_worker_comm_queues()
        self.profiler.stop('_define_worker_comm_queues')
        ###

        ###
        self.profiler.start('_define_dist_graph_comm')
        self._define_dist_graph_comm()
        self.profiler.stop('_define_dist_graph_comm')
        ###

        self._comm_buffer_head = []
        self._recv_points = []
        for r in self.recv_queue.values:
            self._comm_buffer_head.append(np.zeros(len(r)))
            self._recv_points.extend(r)
        self._comm_buffer_flow = np.array(self._comm_buffer_head, dtype = object)
        self._comm_buffer_head = np.array(self._comm_buffer_head, dtype = object)

        ###
        self.profiler.start('_allocate_memory')
        if self.router['main'].rank == 0:
            self.local_to_global = {}
        self._allocate_memory()
        self.profiler.stop('_allocate_memory')
        ###

        ###
        self.profiler.start('_load_initial_conditions')
        self._load_initial_conditions()
        self.profiler.stop('_load_initial_conditions')
        ###

    def _allocate_memory(self):
        self.mem_pool_points = Table2D(MEM_POOL_POINTS, self.num_points, 2)
        self.point_properties = Table(POINT_PROPERTIES, self.num_points)
        if self.num_nodes > 0:
            self.results['node'] = Table2D(NODE_RESULTS, self.num_nodes, self.time_steps,
                labels = self.ss['node'].labels[self.where.nodes['all_to_points',]])

        are_my_uboundaries = self.global_where.points['are_uboundaries'] \
            [self.processors[self.global_where.points['are_uboundaries']] == self.router['main'].rank]
        self.where.points['are_my_uboundaries'] = self.local_points[np.isin(self.points, are_my_uboundaries)]

        are_my_dboundaries = self.global_where.points['are_dboundaries'] \
            [self.processors[self.global_where.points['are_dboundaries']] == self.router['main'].rank]
        self.where.points['are_my_dboundaries'] = self.local_points[np.isin(self.points, are_my_dboundaries)]

        ppoints_start = self.points[self.where.points['are_my_dboundaries']]
        ppoints_end = self.points[self.where.points['are_my_uboundaries']]
        pipes_start = self.global_where.points['to_pipes'][ppoints_start]
        pipes_end = self.global_where.points['to_pipes'][ppoints_end]

        self.num_start_pipes = len(ppoints_start)
        self.num_end_pipes = len(ppoints_end)
        if self.num_start_pipes > 0:
            self.results['pipe.start'] = Table2D(PIPE_START_RESULTS, len(ppoints_start), self.time_steps, labels = self.ss['pipe'].labels[pipes_start])
        if self.num_end_pipes > 0:
            self.results['pipe.end'] = Table2D(PIPE_END_RESULTS, len(ppoints_end), self.time_steps, labels = self.ss['pipe'].labels[pipes_end])

        # Root processor gathers indexes to facilitate reading results

        node_indexes = self.router['main'].gather(self.where.nodes['all_to_points',], root = 0)
        pipe_start_indexes = self.router['main'].gather(pipes_start, root = 0)
        pipe_end_indexes = self.router['main'].gather(pipes_end, root = 0)

        if self.router['main'].rank == 0:
            node_indexes = np.concatenate(node_indexes)
            pipe_start_indexes = np.concatenate(pipe_start_indexes)
            pipe_end_indexes = np.concatenate(pipe_end_indexes)

            node_labels = self.ss['node'].labels[node_indexes]
            pipe_start_labels = self.ss['pipe'].labels[pipe_start_indexes]
            pipe_end_labels = self.ss['pipe'].labels[pipe_end_indexes]

            self.local_to_global['node'] = {l : i for i, l in enumerate(node_labels)}
            self.local_to_global['pipe.start'] = {l : i for i, l in enumerate(pipe_start_labels)}
            self.local_to_global['pipe.end'] = {l : i for i, l in enumerate(pipe_end_labels)}

    def _define_dist_graph_comm(self):
        self.router.add_communicator('local', self.router['main'].Create_dist_graph_adjacent(
            sources = self.recv_queue.labels,
            destinations = self.send_queue.labels,
            sourceweights = list(map(len, self.recv_queue.values)),
            destweights = list(map(len, self.send_queue.values))))

    def _define_worker_comm_queues(self):
        local_points = self.partition['points']['local_idx']
        pp = self.processors[self.points]
        pp_idx = np.where(pp != self.router['main'].rank)[0]
        ppoints = self.points[pp_idx]
        # Define receive queue
        self.recv_queue = ObjArray()
        for i, p in enumerate(pp_idx):
            if not pp[p] in self.recv_queue.indexes:
                self.recv_queue[pp[p]] = []
            self.recv_queue[pp[p]].append(ppoints[i])

        # Define send queue
        self.send_queue = ObjArray()
        uboundaries = self.points[self.where.points['are_uboundaries']]
        dboundaries = self.points[self.where.points['are_dboundaries']]
        inner = self.points[self.where.points['are_inner']]

        for p in self.recv_queue.labels:
            self.recv_queue[p] = np.sort(self.recv_queue[p])
            urq = np.isin(self.recv_queue[p], uboundaries)
            drq = np.isin(self.recv_queue[p], dboundaries)
            irq = np.isin(self.recv_queue[p], inner)

            extra_b = np.append(self.recv_queue[p][urq] - 1, self.recv_queue[p][drq] + 1)
            extra_i = np.append(self.recv_queue[p][irq] - 1, self.recv_queue[p][irq] + 1)
            extra = np.append(extra_b, extra_i)
            reduced_extra = extra[np.isin(extra, self.points)]
            real_extra = [local_points[r] for r in reduced_extra[self.processors[reduced_extra] == self.router['main'].rank]] # local idx
            if len(real_extra) > 0:
                if not p in self.send_queue.indexes:
                    self.send_queue[p] = []
                self.send_queue[p].extend(real_extra)
            self.recv_queue[p] = np.sort([local_points[r] for r in self.recv_queue[p]]) # convert to local idx

        for p in self.send_queue.labels:
            self.send_queue[p] = np.sort(np.unique(self.send_queue[p]))

    def _create_selectors(self):
        jip_nodes = self.partition['nodes']['global_idx']
        lpoints = self.partition['points']['local_idx']

        self.where.points['just_in_pipes'] = np.array([lpoints[npoint] for npoint in self.partition['nodes']['points']]).astype(int)
        self.where.points['are_tanks'] = np.where(np.isin(self.points, self.partition['tanks']['points']))[0]
        self.where.points['are_reservoirs'] = np.where(np.isin(self.points, self.partition['reservoirs']['points']))[0]
        njip = np.cumsum(self.partition['nodes']['context'])
        self.where.nodes['just_in_pipes',] = njip[:-1]
        self.where.nodes['to_points'] = self.where.points['just_in_pipes'][self.where.nodes['just_in_pipes',][:-1]]

        nonpipe = np.isin(self.global_where.points['are_boundaries'], self.global_where.points['are_valve'])
        nonpipe = nonpipe | np.isin(self.global_where.points['are_boundaries'], self.global_where.points['are_pump'])
        local_points = np.isin(self.global_where.points['are_boundaries'], self.points[self.processors[self.points] == self.router['main'].rank])
        dboundary = np.zeros(len(nonpipe), dtype=bool); dboundary[::2] = 1
        uboundary = np.zeros(len(nonpipe), dtype=bool); uboundary[1::2] = 1
        # ---------------------------
        self.where.points['are_uboundaries'] = np.where(np.isin(self.points, self.global_where.points['are_uboundaries']))[0]
        self.where.points['are_dboundaries'] = np.where(np.isin(self.points, self.global_where.points['are_dboundaries']))[0]
        self.where.points['are_inner'] = np.setdiff1d(np.arange(self.num_points, dtype=np.int), \
            np.concatenate((self.where.points['are_uboundaries'], self.where.points['are_dboundaries'])))
        # ---------------------------
        n_pipes = len(self.global_where.points['are_uboundaries'])
        ppipes_idx = np.arange(n_pipes, dtype=int)
        ppipes = np.zeros(n_pipes*2, dtype=int)
        ppipes[::2] = ppipes_idx; ppipes[1::2] = ppipes_idx
        selector_dboundaries = dboundary & (~nonpipe) & local_points
        self.where.points['jip_dboundaries'] = np.where(np.isin(self.points, self.global_where.points['are_boundaries'][selector_dboundaries]))[0]
        self.where.points['jip_dboundaries',] = ppipes[selector_dboundaries]
        selector_uboundaries = uboundary & (~nonpipe) & local_points
        self.where.points['jip_uboundaries'] = np.where(np.isin(self.points, self.global_where.points['are_boundaries'][selector_uboundaries]))[0]
        self.where.points['jip_uboundaries',] = ppipes[selector_uboundaries]
        # ---------------------------
        diff = np.diff(njip)
        self.where.points['just_in_pipes',] = np.array([i for i in range(len(jip_nodes)) for j in range(diff[i])], dtype = int)
        # ---------------------------
        self.where.points['start_valve'] = np.array([lpoints[spv] for spv in self.partition['inline_valves']['start_points']]).astype(int)
        self.where.points['end_valve'] = np.array([lpoints[epv] for epv in self.partition['inline_valves']['end_points']]).astype(int)
        self.where.points['start_valve',] = self.partition['inline_valves']['global_idx']
        self.where.points['start_pump'] = np.array([lpoints[spp] for spp in self.partition['inline_pumps']['start_points']]).astype(int)
        self.where.points['end_pump'] = np.array([lpoints[epv] for epv in self.partition['inline_pumps']['end_points']]).astype(int)
        self.where.points['start_pump',] = self.partition['inline_pumps']['global_idx']
        self.where.points['single_valve'] = np.array([lpoints[svp] for svp in self.partition['single_valves']['points']]).astype(int)
        self.where.points['single_valve',] = self.partition['single_valves']['global_idx']
        self.where.points['single_pump'] = np.array([lpoints[spp] for spp in self.partition['single_pumps']['points']]).astype(int)
        self.where.points['single_pump',] = self.partition['single_pumps']['global_idx']
        # ---------------------------
        nodes = []; node_points = []
        nodes += list(self.partition['nodes']['global_idx'])
        node_points += list(self.partition['nodes']['points'][self.where.nodes['just_in_pipes',]])
        nodes += list(self.partition['tanks']['global_idx'])
        node_points += list(self.partition['tanks']['points'])
        nodes += list(self.partition['reservoirs']['global_idx'])
        node_points += list(self.partition['reservoirs']['points'])
        nodes += list(self.ss['valve'].start_node[self.partition['inline_valves']['global_idx']])
        node_points += list(self.partition['inline_valves']['start_points'])
        nodes += list(self.ss['valve'].end_node[self.partition['inline_valves']['global_idx']])
        node_points += list(self.partition['inline_valves']['end_points'])
        nodes += list(self.ss['pump'].start_node[self.partition['inline_pumps']['global_idx']])
        node_points += list(self.partition['inline_pumps']['start_points'])
        nodes += list(self.ss['pump'].end_node[self.partition['inline_pumps']['global_idx']])
        node_points += list(self.partition['inline_pumps']['end_points'])
        nodes += list(self.ss['valve'].start_node[self.partition['single_valves']['global_idx']])
        node_points += list(self.partition['single_valves']['points'])
        nodes += list(self.ss['pump'].end_node[self.partition['single_pumps']['global_idx']])
        node_points += list(self.partition['single_pumps']['points'])
        nodes = np.array(nodes)
        node_points = np.array(node_points)
        if len(nodes) > 0:
            atp = np.array([lpoints[npoint] for npoint in node_points]).astype(int)
            _, idx_unique = np.unique(nodes, return_index=True)
            sorted_idx = np.sort(idx_unique)
            self.where.nodes['all_to_points'] = atp[sorted_idx]
            self.where.nodes['all_to_points',] = nodes[sorted_idx]
            self.num_nodes = len(self.where.nodes['all_to_points',])
            self.where.nodes['all_just_in_pipes'] = self.partition['nodes']['global_idx']
            self.num_jip_nodes = len(self.where.nodes['all_just_in_pipes'])
        # ---------------------------
        if self.ss['open_protection']:
            ssprotection = np.isin(self.partition['nodes']['points'], self.global_where.points['start_open_protection'])
            esprotection = np.isin(self.partition['nodes']['points'], self.global_where.points['end_open_protection'])
            self.where.points['start_open_protection'] = np.array([lpoints[npoint] for npoint in self.partition['nodes']['points'][ssprotection]]).astype(int)
            self.where.points['end_open_protection'] = np.array([lpoints[npoint] for npoint in self.partition['nodes']['points'][esprotection]]).astype(int)
            self.num_open_protections = len(self.where.points['start_open_protection'])
        if self.ss['closed_protection']:
            ssprotection = np.isin(self.partition['nodes']['points'], self.global_where.points['start_closed_protection'])
            esprotection = np.isin(self.partition['nodes']['points'], self.global_where.points['end_closed_protection'])
            self.where.points['start_closed_protection'] = np.array([lpoints[npoint] for npoint in self.partition['nodes']['points'][ssprotection]]).astype(int)
            self.where.points['end_closed_protection'] = np.array([lpoints[npoint] for npoint in self.partition['nodes']['points'][esprotection]]).astype(int)
            self.num_closed_protections = len(self.where.points['start_closed_protection'])

    def _fix_protection_indexes(self, protection_type):
        if (self.ss[f'{protection_type}_protection']):
            for ii in range(self.ss[f'{protection_type}_protection'].shape[1]):
                p1 = self.where.points[f'start_{protection_type}_protection']
                p2 = self.where.points[f'end_{protection_type}_protection']
                if self.point_properties.has_plus[p1] == 0: # p1 is not associated with plus characteristic
                    self.where.points[f'end_{protection_type}_protection'] = p1
                    self.where.points[f'start_{protection_type}_protection'] = p2

    def define_initial_conditions_for_points(self, points, pipe, start, end):
        q = self.ss['pipe'].flowrate[pipe]
        self.mem_pool_points.flowrate[start:end,0] = q

        start_node = self.ss['pipe'].start_node[pipe]
        start_point = self.global_where.points['are_boundaries'][pipe*2]
        npoints = points - start_point # normalized

        shead = self.ss['node'].head[start_node]
        self.point_properties.B[start:end] = self.ss['pipe'].wave_speed[pipe] / (G * self.ss['pipe'].area[pipe])
        self.point_properties.R[start:end] = self.ss['pipe'].ffactor[pipe] * self.ss['pipe'].dx[pipe] / \
            (2 * G * self.ss['pipe'].diameter[pipe] * self.ss['pipe'].area[pipe] ** 2)
        per_unit_hl = self.ss['pipe'].head_loss[pipe] / self.ss['pipe'].segments[pipe]
        self.mem_pool_points.head[start:end,0] = shead - per_unit_hl*npoints

    def _load_initial_conditions(self):
        points = self.partition['points']['global_idx']
        pipes = self.global_where.points['to_pipes'][points]
        diff = np.where(np.diff(pipes) >= 1)[0] + 1
        if len(diff) > 0:
            for i in range(len(diff)+1):
                if i == 0:
                    start = 0
                    end = diff[i]
                elif i == len(diff):
                    start = diff[i-1]
                    end = None
                else:
                    start = diff[i-1]
                    end = diff[i]
                self.define_initial_conditions_for_points(points[start:end], pipes[start], start, end)
        else:
            self.define_initial_conditions_for_points(points, pipes[0], 0, None)

        self.point_properties.has_plus[self.where.points['are_uboundaries']] = 1
        self.point_properties.has_minus[self.where.points['are_dboundaries']] = 1
        self.point_properties.has_plus[self.where.points['are_inner']] = 1
        self.point_properties.has_minus[self.where.points['are_inner']] = 1

        self._fix_protection_indexes('open')
        self._fix_protection_indexes('closed')

        if self.num_start_pipes > 0:
            self.results['pipe.start'].flowrate[:,0] = self.mem_pool_points.flowrate[self.where.points['are_my_dboundaries'], 0]
        if self.num_end_pipes > 0:
            self.results['pipe.end'].flowrate[:,0] = self.mem_pool_points.flowrate[self.where.points['are_my_uboundaries'], 0]
        if self.num_nodes > 0:
            self.results['node'].head[:, 0] = self.mem_pool_points.head[self.where.nodes['all_to_points'], 0]
            self.results['node'].leak_flow[:, 0] = \
                self.ss['node'].leak_coefficient[self.where.nodes['all_to_points',]] * \
                    np.sqrt(self.ss['node'].pressure[self.where.nodes['all_to_points',]])
            self.results['node'].demand_flow[:, 0] = \
                self.ss['node'].demand_coefficient[self.where.nodes['all_to_points',]] * \
                    np.sqrt(self.ss['node'].pressure[self.where.nodes['all_to_points',]])

        # Define initial conditions for surge protections
        if self.num_closed_protections > 0:
            H0 = self.mem_pool_points.head[:,0]
            m = 1.2
            Hb = 10.3 # barometric pressure [mH2O]
            self.ss['closed_protection'].HT0[:] = self.ss['closed_protection'].water_level
            self.ss['closed_protection'].HA[:] = self.ss['node'].head[self.ss['closed_protection'].node] - self.ss['closed_protection'].water_level + Hb # air pressure head
            self.ss['closed_protection'].VA[:] = self.ss['closed_protection'].area*(self.ss['closed_protection'].height-self.ss['closed_protection'].water_level) # air volume
            self.ss['closed_protection'].C[:] = self.ss['closed_protection'].HA * self.ss['closed_protection'].VA**m
            # Allocate space for results
            # self.results['closed_protection'] = Table2D(CLOSED_PROTECTION_RESULTS, self.num_closed_protections, self.time_steps, self)


    def exchange_data(self, t):
        t1 = t % 2; t0 = 1 - t1
        send_flow = []
        send_head = []
        for v in self.send_queue.values:
            send_flow.append(self.mem_pool_points.flowrate[v,t1])
            send_head.append(self.mem_pool_points.head[v,t1])
        self._comm_buffer_flow = self.router['local'].neighbor_alltoall(send_flow)
        self._comm_buffer_head = self.router['local'].neighbor_alltoall(send_head)
        self.mem_pool_points.flowrate[self._recv_points, t1] = [item for sublist in self._comm_buffer_flow for item in sublist]
        self.mem_pool_points.head[self._recv_points, t1] = [item for sublist in self._comm_buffer_head for item in sublist]

    def run_step(self, t):
        t1 = t % 2; t0 = 1 - t1

        Q0 = self.mem_pool_points.flowrate[:,t0]
        H0 = self.mem_pool_points.head[:,t0]
        Q1 = self.mem_pool_points.flowrate[:,t1]
        H1 = self.mem_pool_points.head[:,t1]

        ###
        self.profiler.start('run_interior_step')
        run_interior_step(
            Q0, H0, Q1, H1,
            self.point_properties.B,
            self.point_properties.R,
            self.point_properties.Cp,
            self.point_properties.Bp,
            self.point_properties.Cm,
            self.point_properties.Bm,
            self.point_properties.has_plus,
            self.point_properties.has_minus)
        self.profiler.stop('run_interior_step')
        ###

        ###
        self.profiler.start('run_general_junction')
        if 'node' in self.results: # worker has junctions
            run_general_junction(
                H0, Q1, H1,
                self.results['node'].leak_flow[:,t],
                self.results['node'].demand_flow[:,t],
                self.point_properties.Cp,
                self.point_properties.Bp,
                self.point_properties.Cm,
                self.point_properties.Bm,
                self.ss['node'].leak_coefficient,
                self.ss['node'].demand_coefficient,
                self.ss['node'].elevation,
                self.where)
        self.profiler.stop('run_general_junction')
        ###

        ###
        self.profiler.start('run_valve_step')
        run_valve_step(
            Q1, H1,
            self.point_properties.Cp,
            self.point_properties.Bp,
            self.point_properties.Cm,
            self.point_properties.Bm,
            self.ss['valve'].setting,
            self.ss['valve'].K,
            self.ss['valve'].area,
            self.where)
        self.profiler.stop('run_valve_step')
        ###

        ###
        self.profiler.start('run_pump_step')
        run_pump_step(
            self.ss['pump'].source_head,
            Q1, H1,
            self.point_properties.Cp,
            self.point_properties.Bp,
            self.point_properties.Cm,
            self.point_properties.Bm,
            self.ss['pump'].a1,
            self.ss['pump'].a2,
            self.ss['pump'].Hs,
            self.ss['pump'].setting,
            self.where)
        self.profiler.stop('run_pump_step')
        ###

        ###
        if self.num_open_protections > 0:
            self.profiler.start('run_open_protections')
            run_open_protections(
                H0, H1,
                Q1,
                self.ss['open_protection'].QT,
                self.ss['open_protection'].area,
                self.point_properties.Cp,
                self.point_properties.Bp,
                self.point_properties.Cm,
                self.point_properties.Bm, self.time_step, self.where)
            self.profiler.stop('run_open_protections')
        ###

        ###
        if self.num_closed_protections > 0:
            self.profiler.start('run_closed_protections')
            run_closed_protections(
                H0, H1,
                Q1,
                self.ss['closed_protection'].QT0,
                self.ss['closed_protection'].QT1,
                self.ss['closed_protection'].HT0,
                self.ss['closed_protection'].HT1,
                self.ss['closed_protection'].VA,
                self.ss['closed_protection'].height,
                self.ss['closed_protection'].area,
                self.point_properties.Cp,
                self.point_properties.Bp,
                self.point_properties.Cm,
                self.point_properties.Bm, self.time_step,
                self.ss['closed_protection'].C, self.where)
            self.profiler.stop('run_closed_protections')
        ###

        self.profiler.start('store_results')
        if self.num_start_pipes > 0:
            self.results['pipe.start'].flowrate[:,t] = self.mem_pool_points.flowrate[self.where.points['are_my_dboundaries'], t1]
        if self.num_end_pipes > 0:
            self.results['pipe.end'].flowrate[:,t] = self.mem_pool_points.flowrate[self.where.points['are_my_uboundaries'], t1]
        if self.num_nodes > 0:
            self.results['node'].head[:, t] = self.mem_pool_points.head[self.where.nodes['all_to_points'], t1]
        self.profiler.stop('store_results')
        ###





import shutil
import os
import pickle
import pathlib

from datetime import datetime
from ptsnet.utils.io import get_root_path, walk
from pkg_resources import resource_filename
from ptsnet.results.storage import StorageManager
from ptsnet.utils.data import is_array

def delete_all_workspaces():
    confirmation = None
    while confirmation is None:
        confirmation = input('Are you sure that you want to delete all the workspaces? (yes/no): ')
        if not confirmation in ('yes', 'no'):
            confirmation = None
    if confirmation == 'no':
        return
    ROOT = get_root_path()
    for d in os.listdir(os.path.join(ROOT, 'workspaces')):
        abs_path = os.path.join(get_root_path(), 'workspaces', d)
        if os.path.isdir(abs_path):
            shutil.rmtree(abs_path)
        else:
            os.remove(abs_path)

def delete_workspace(workspace_id, full_path=True):
    confirmation = None

    ROOT = get_root_path()
    wpl = list_workspaces()
    widlist = workspace_id

    txt = {}

    wids = workspace_id
    if not is_array(workspace_id):
        wids = [workspace_id]
        apaths = [os.path.join(get_root_path(), 'workspaces', wpl[workspace_id])]
    else:
        apaths = \
            [os.path.join(
                get_root_path(),
                'workspaces',
                wpl[wp_id]) for wp_id in workspace_id]

    for i, apath in enumerate(apaths):
        f = pathlib.Path(apath)
        st = f.stat().st_mtime
        with open(os.path.join(apath, 'settings.pkl'), 'rb') as f:
            s = pickle.load(f)
        with open(os.path.join(apath, 'fname.pkl'), 'rb') as f:
            fname = pickle.load(f)
        if not full_path:
            fname = os.path.basename(fname)
        txt = [
            f'Last modification on: {datetime.fromtimestamp(st)}',
            fname,
            f"T = {s['duration']}, t = {s['time_step']}, N = {s['num_points']}, n = {s['num_processors']}",
        ]

        print('\n')
        print(f'  ({wids[i]}) ' + txt[0])
        print('      ' + txt[1])
        print('      ' + txt[2])
        print('\n')

        confirmation = None
        while confirmation is None:
            confirmation = input('Are you sure that you want to delete this workspace ? (yes/no): ')
            if not confirmation in ('yes', 'no'):
                confirmation = None

        if confirmation == 'no':
            continue
        if os.path.isdir(apath):
            shutil.rmtree(apath)
        else:
            os.remove(apath)

def list_workspaces():
    wps = [d for d in os.listdir(os.path.join(get_root_path(), 'workspaces')) if '.' not in d]
    # workspace code starts with 'W' and then is followed by a number
    wps.sort(key = lambda x: int(x[1:]))
    return wps

def print_workspaces(full_path = False):
    txt = {}
    for d in list_workspaces():
        abs_path = os.path.join(get_root_path(), 'workspaces', d)
        with open(os.path.join(abs_path, 'settings.pkl'), 'rb') as f:
            s = pickle.load(f)
        with open(os.path.join(abs_path, 'fname.pkl'), 'rb') as f:
            fname = pickle.load(f)
            if full_path:
                fname = os.path.basename(fname)
        txt[d] = [
            fname,
            f"T = {s['duration']}, t = {s['time_step']}, N = {s['num_points']}, n = {s['num_processors']}",
        ]
    print('\n')
    for i, d in enumerate(txt):
        print(f'  ({i})', txt[d][0])
        print('    ' + txt[d][1])
        print('\n')

def num_workspaces():
    return len(list_workspaces())

def get_count_path():
    return os.path.join(get_root_path(), 'workspaces', 'count.pkl')

def new_workspace_name(is_root = True):
    if is_root:
        if not os.path.exists(get_count_path()):
            count = 0
        else:
            with open(get_count_path(), 'rb') as f:
                count = pickle.load(f)
        count += 1
        with open(get_count_path(), 'wb') as f:
            pickle.dump(count, f)
        return f'W{count}'
    return None

def exists(data_label, workspace_id=None, workspace_name=None):
    if workspace_id:
        wps = list_workspaces()
        stmanager = StorageManager(wps[workspace_id])
    elif workspace_name:
        stmanager = StorageManager(workspace_name)
    else:
        raise ValueError("It is necessary to define either workspace_id or workspace_name")
    return stmanager.exists(data_label)